from __future__ import annotations
from typing import TYPE_CHECKING
import numpy as np
from numbers import Number
import scipy.stats as sps
import pyuncertainnumber.pba.intervals.intervalOperators as io
from ..pba.intervals import Interval
from pyuncertainnumber import pba
from ..pba.distributions import Distribution as D
from ..pba.distributions import named_dists
from ..pba.pbox_parametric import named_pbox
import functools
from .core import makeUN

if TYPE_CHECKING:
    from pyuncertainnumber import Interval
    from pyuncertainnumber import UncertainNumber

""" statistical inference functions from data for the UncertainNumber class. """


@makeUN
def fit(method: str, family: str, data: np.ndarray) -> UncertainNumber:
    """parametric estimator to fit a distribution from data

    args:
        method (str): method of fitting, e.g., {'mle' or 'mom'} 'entropy', 'pert', 'fermi', 'bayesian'

        family (str): distribution family to be fitted

        data (np.ndarray): data to be fitted


    return:
        - UncertainNumber object

    example:
        >>> # precise data
        >>> pun.fit('mle', 'norm', np.random.normal(0, 1, 100))
        >>>  # imprecise data
        >>> precise_sample = sps.expon(scale=1/0.4).rvs(15)
        >>> imprecise_data = pba.I(lo = precise_sample - 1.4, hi=precise_sample + 1.4)
        >>> pun.fit('mom', family='exponential', data=imprecise_data)


    .. seealso::
        :func:`pyuncertainnumber.pba.KS_bounds` : a non-parametric charactearisation method using Kolmogorov-Smirnov bounds
    """
    match method:
        case "mle":
            try:
                return named_mle_methods.get(family)(data)
            except:
                return named_smle_methods.get(family)(data)
        case "mom":
            return named_mom_methods.get(family)(data)
        case _:
            raise ValueError("method not supported")


def makedist(shape: str):
    """a decorator to transform a `sps.dist` into `Distribution` objects"""

    def decorator_make_dist(func):
        @functools.wraps(func)
        def wrapper_decorator(*args, **kwargs):  # input array x
            return D.dist_from_sps(func(*args, **kwargs), shape)

        return wrapper_decorator

    return decorator_make_dist


###############################################################################
# Method-of-Moment distribution constructors (matching central moments of x)
###############################################################################
""" hint: given interval data x, return pbox """


def MMbernoulli(x):
    """a first attempt to Maximum likelihood estimation for Bernoulli distribution which accepts both precise and imprecise data;

    #! the example of `singleparam` pattern
    #! to change, add the 'interval_measurement' decorator
    note:
        the attempt is successful per se, but not accommodating to the top-level calling signature yet.

        - precise data returns precise distrubution
        - imprecise data need to be in Interval type to return a pbox
        - interval data can return either a precise distribution or a pbox
    """
    if isinstance(x, sps.CensoredData | np.ndarray | list):
        return D.dist_from_sps(sps.expon(*sps.bernoulli.fit(x)), shape="bernoulli")
    elif isinstance(x, Interval):
        return pba.bernoulli(io.mean(x))
    else:
        raise TypeError("Input data type not supported")


@makedist("beta")
def MMbeta(x: np.ndarray):
    m = x.mean()
    s = x.var()

    alpha = m * (m * (1 - m) / s - 1)
    beta_p = (1 - m) * (m * (1 - m) / s - 1)
    return sps.beta(alpha, beta_p)


@makedist("beta")
def mm_beta(mean, std):
    """from given moments, return a beta distribution"""
    m = mean
    s = std**2

    alpha = m * (m * (1 - m) / s - 1)
    beta_p = (1 - m) * (m * (1 - m) / s - 1)
    return sps.beta(alpha, beta_p)


@makedist("betabinom")
def MMbetabinomial(n: int, x):  # **
    #! 'x**2' variable repetition
    # n must be provided; it's not estimated from data
    # https://en.wikipedia.org/wiki/Beta-binomial_distribution#Example:
    # MMbetabinomial(n=12,rep(0:12,c(3,24,104,286,670,1033,1343,1112,829,478,181,45,7)))
    m1 = x.mean()
    m2 = (x**2).mean()
    d = n * (m2 / m1 - m1 - 1) + m1
    return sps.betabinom(n, (n * m1 - m2) / d, (n - m1) * (n - m2 / m1) / d)


@makedist("binom")
def MMbinomial(x, n: int):  # **
    #! the return seems overcomplicated?
    """
    args:
         - n (int): number of trials
    """
    return sps.binom(n, io.mean() / n)
    # a = x.mean()
    # b = x.std()
    # return binom(int(np.abs(np.round(a/(1-b**2/a)))), np.abs(1-b**2/a))


@makedist("chi2")
def MMchisquared(x):
    return sps.chi2(np.round(x.mean()))


# def MMchisquared(x):
#     #! TODO interval outward_rounding
#     if isinstance(x, sps.CensoredData | np.ndarray | list):
#         return D.dist_from_sps(chi2(outward_rounding(mean(x))), shape='chi2')
#     elif isinstance(x, Interval):
#         return pba.expon(mean(x))
#     else:
#         raise TypeError('Input data type not supported')


def MMexponential(x):
    import pyuncertainnumber.pba.intervals.intervalOperators as io

    mean_ = io.mean(x)
    return universal_exponential(mean_)


def universal_exponential(mean):
    """single parameter pattern exponential distribution constructor

    note:
        mean can be a precise number or an interval; Exponential is parameterised by scale, which is the inverse of the rate parameter (lambda).
    """
    if isinstance(mean, Number):
        return mm_exponential(mean)
    elif isinstance(mean, Interval):
        return pba.exponential(scale=mean)


@makedist("exponential")
def mm_exponential(mean):
    """from given moments, return an exponential distribution

    note:
        This constructor is parameterised by scale, which is the inverse of the rate parameter (lambda).
    """
    return sps.expon(scale=mean)  # scale is 1/lambda


@makedist("f")
def MMF(x):  # **
    w = 2 / (1 - 1 / x.mean())
    return sps.f(
        np.round(
            (2 * w**3 - 4 * w**2) / ((w - 2) ** 2 * (w - 4) * x.std() ** 2 - 2 * w**2)
        ),
        np.round(w),
    )


@makedist("gamma")
def MMgamma(x):  # **
    a = x.mean()
    b = x.std()
    return sps.gamma(b**2 / a, 1 / (a / b) ** 2)  # gamma1(a, b) ~ gamma(b²/a, (a/b)²)


@makedist("gamma")
def mm_gamma(mean, std=None):
    """from given moments, return a gamma distribution"""

    return sps.gamma(
        std**2 / mean, 1 / (mean / std) ** 2
    )  # gamma1(a, b) ~ gamma(b²/a, (a/b)²)


@makedist("geometric")
def MMgeometric(x):
    pass


def MMgeometric(x):
    if isinstance(x, sps.CensoredData | np.ndarray | list):
        return D.dist_from_sps(sps.geom(1 / (1 + io.mean(x))), shape="geometric")
    elif isinstance(x, Interval):
        return pba.geom(1 / (1 + io.mean(x)))
    else:
        raise TypeError("Input data type not supported")


def MMpascal(x):
    if isinstance(x, sps.CensoredData | np.ndarray | list):
        return D.dist_from_sps(sps.geom(1 / (1 + io.mean(x))), shape="pascal")
    elif isinstance(x, Interval):
        return pba.geom(1 / (1 + io.mean(x)))
    else:
        raise TypeError("Input data type not supported")


# def MMgumbel0(x) : return(gumbel(x.mean() - 0.57721 * x.std() * np.sqrt(6)/ np.pi, x.std() * np.sqrt(6)/ np.pi))       #**  # https://stackoverflow.com/questions/51427764/using-method-of-moments-with-gumbel-r-in-python-scipy-stats-gumbel-r


@makedist("gumbel")
def MMgumbel(x):  # **
    # https://stackoverflow.com/questions/51427764/using-method-of-moments-with-gumbel-r-in-python-scipy-stats-gumbel-r
    scale = np.sqrt(6) / np.pi * np.std(x)
    loc = np.mean(x) - np.euler_gamma * scale
    return sps.gumbel_r(loc, scale)


@makedist("gumbel")
def mm_gumbel(mean, std):
    """from given moments, return a gumbel distribution"""
    scale = np.sqrt(6) / np.pi * std
    loc = mean - np.euler_gamma * scale
    return sps.gumbel_r(loc, scale)


@makedist("extremevalue")
def MMextremevalue(x):
    return sps.gumbel_r(
        x.mean() - 0.57721 * x.std() * np.sqrt(6) / np.pi, x.std() * np.sqrt(6) / np.pi
    )  # **


# TODO possibly incorrect. See the function below.
@makedist("lognormal")
def MMlognormal(x):
    return sps.lognorm(x.mean(), x.std())  # **


def lognormal_mom_estimator(data=None, sample_mean=None, sample_variance=None):
    """
    Method of Moments estimator for Lognormal(μ, σ^2).

    Parameters
    ----------
    data : array-like, optional
        Raw data sample. If provided, mean and variance are computed from it.
    sample_mean : float, optional
        Precomputed sample mean (ignored if data is provided).
    sample_variance : float, optional
        Precomputed sample variance (ignored if data is provided).

    Returns
    -------
    mu_hat : float
        Estimated mean of the underlying normal distribution.
    sigma2_hat : float
        Estimated variance of the underlying normal distribution.
    """
    # If raw data is provided, compute mean and variance
    if data is not None:
        data = np.asarray(data)
        if np.any(data <= 0):
            raise ValueError(
                "All data points must be positive for lognormal distribution."
            )
        sample_mean = data.mean()
        sample_variance = data.var(ddof=0)  # population variance
    elif sample_mean is None or sample_variance is None:
        raise ValueError(
            "Provide either raw data or both sample_mean and sample_variance."
        )

    if sample_mean <= 0:
        raise ValueError("Sample mean must be positive for lognormal distribution.")

    # Method of moments formulas
    sigma2_hat = np.log(1 + sample_variance / (sample_mean**2))
    mu_hat = np.log(sample_mean) - sigma2_hat / 2

    return mu_hat, sigma2_hat


def mm_lognormal(mean: float, std: float):
    """
    Method of Moments estimator for Lognormal(μ, σ^2).

    args
        mean (float):
            Sample mean of the data (m1).
        std (float):
            Sample standard deviation of the data (s).

    returns
        mu_hat (float):
            Estimated mean of the underlying normal distribution.
        sigma2_hat (float):
            Estimated variance of the underlying normal distribution.
    """

    var = std**2

    if mean <= 0:
        raise ValueError("Sample mean must be positive for lognormal distribution.")

    # Estimate sigma^2
    sigma2_hat = np.log(1 + var / (mean**2))

    # Estimate mu
    mu_hat = np.log(mean) - sigma2_hat / 2

    return mu_hat, sigma2_hat


@makedist("laplace")
def MMlaplace(x):
    return sps.laplace(x.mean(), x.std() / np.sqrt(2))  # **


@makedist("doubleexponential")
def MMdoubleexponential(x):
    return sps.laplace(x.mean(), x.std() / np.sqrt(2))  # **


@makedist("logistic")
def MMlogistic(x):
    return sps.logistic(x.mean(), x.std() * np.sqrt(3) / np.pi)  # **


@makedist("loguniform")
def MMloguniform(x):
    return sps.loguniform(mean=x.mean(), std=x.std())  # **


@makedist("norm")
def MMnormal(x):
    return sps.norm(x.mean(), x.std())  # **


@makedist("norm")
def mm_normal(mean, std):
    """from given moments, return a normal distribution"""

    return sps.norm(loc=mean, scale=std)


@makedist("gaussian")
def MMgaussian(x):
    return sps.norm(x.mean(), x.std())  # **


@makedist("pareto")
def MMpareto(x):  # **
    a = x.mean()
    b = x.std()
    return sps.pareto(
        a / (1 + 1 / np.sqrt(1 + a**2 / b**2)), 1 + np.sqrt(1 + a**2 / b**2)
    )


def MMpoisson(x):
    shape = "poisson"
    if isinstance(x, sps.CensoredData | np.ndarray | list):
        return D.dist_from_sps(named_dists.get(shape)(io.mean(x)), shape=shape)
    elif isinstance(x, Interval):
        return named_pbox.get(shape)(io.mean(x))
    else:
        raise TypeError("Input data type not supported")


@makedist("powerlaw")
def MMpowerfunction(x):  # **
    a = x.mean()
    b = x.std()
    return sps.powerlaw(
        a / (1 - 1 / np.sqrt(1 + (a / b) ** 2)), np.sqrt(1 + (a / b) ** 2) - 1
    )


@makedist("t")
def MMt(x):
    return sps.t(2 / (1 - 1 / x.std() ** 2))


@makedist("student")
def MMstudent(x):
    assert not (1 < x.std()), "Improper standard deviation for student distribution"
    return sps.t(2 / (1 - 1 / x.std() ** 2))


@makedist("uniform")
def MMuniform(x):  # **
    a = x.mean()
    b = np.sqrt(3) * x.std()
    return sps.uniform(a - b, a + b)


@makedist("rectangular")
def MMrectangular(x):
    return MMuniform(x)


@makedist("triangular")
def MMtriangular(x, iters=100, dives=10):  # **
    # iterative search for triangular distribution parameters using method of
    # matching moments (you solve the thing analytically! too messy without help)
    # testing code indicated with #-#
    # -#some = 10
    # -#A = runif(1,0,10)
    # -#B = A + runif(1,0,10)
    # -#C = runif(1,A,B)
    # -#x = qtriangular(runif(some), A,C,B)
    def skewness(x):
        m = x.mean()
        # std uses the population formula, may need the sample formula
        return np.sum((x - m) ** 3) / ((len(x) - 1) * x.std() ** 3)

    M = np.mean(x)
    V = np.var(x)
    S = skewness(x)
    a = aa = min(x)  # apparently double assignments work
    b = bb = max(x)
    c = cc = 3 * M - a - b
    many = iters
    s1 = np.std(x)
    for k in range(dives):
        s1 = s2 = s3 = s1 / 2
        a = np.random.normal(aa, s1, many)
        b = np.random.normal(bb, s2, many)
        c = np.random.normal(cc, s3, many)
        m = (a + b + c) / 3
        k = a**2 + b**2 + c**2 - a * b - a * c - b * c
        v = k / 18
        s = (np.sqrt(2) * (a + b - 2 * c) * (2 * a - b - c) * (a - 2 * b + c)) / (
            5 * k ** (3 / 2)
        )
        d = (M - m) ** 2 + (V - v) ** 2 + (S - s) ** 2
        i = np.argmin(d)  # which.min(d)
        aa = a[i]
        bb = b[i]
        cc = c[i]
    # -#gray(triangular(A,B,C), new = TRUE)
    # -#blue(x)
    # -#green(triangular(aa,bb,cc))
    # -#A;aa; B;bb; C;cc  # the order is min, max, mode
    print(aa, bb, cc)
    return sps.triang(aa, cc, bb)


# no decorator needed herein
def MMrayleigh(x):
    return mm_rayleigh(x.mean())


def universal_rayleigh(mean):
    """single parameter pattern rayleigh distribution constructor

    note:
        mean can be a precise number or an interval;
    """
    if isinstance(mean, Number):
        return mm_rayleigh(mean)
    elif isinstance(mean, Interval):
        _scale = mean / np.sqrt(np.pi / 2)
        return pba.rayleigh(scale=_scale)


@makedist("rayleigh")
def mm_rayleigh(mean):
    """from given moments, return a rayleigh distribution"""
    return sps.rayleigh(scale=mean / np.sqrt(np.pi / 2))


named_mom_methods = {
    "bernoulli": MMbernoulli,
    "beta": MMbeta,
    "betabinomial": MMbetabinomial,
    "binomial": MMbinomial,
    "chisquared": MMchisquared,
    "exponential": MMexponential,
    "expon": MMexponential,
    "F": MMF,
    "f": MMF,
    "gamma": MMgamma,
    "geometric": MMgeometric,
    "gumbel": MMgumbel,
    "extremevalue": MMextremevalue,
    "lognormal": MMlognormal,
    "laplace": MMlaplace,
    "doubleexponential": MMdoubleexponential,
    "logistic": MMlogistic,
    "loguniform": MMloguniform,
    "norm": MMnormal,
    "normal": MMnormal,
    "gaussian": MMgaussian,
    "pareto": MMpareto,
    "poisson": MMpoisson,
    "powerfunction": MMpowerfunction,
    "rayleigh": MMrayleigh,
    "t": MMt,
    "student": MMstudent,
    "uniform": MMuniform,
    "rectangular": MMrectangular,
    "triangular": MMtriangular,
}

###############################################################################
# * Alternative maximum likelihood estimation constructors using scipy.stats  *#
###############################################################################

# Some of these functions may support intervals in the data x.  See
# https://docs.scipy.org/doc/scipy/reference/generated/scipy.stats.rv_continuous.fit.html#scipy.stats.rv_continuous.fit
# https://docs.scipy.org/doc/scipy/reference/generated/scipy.stats.CensoredData.html#scipy.stats.CensoredData


@makedist("bernoulli")
def MLbernoulli(x):
    return sps.bernoulli(*sps.bernoulli.fit(x))


@makedist("beta")
def MLbeta(x):
    return sps.beta(*sps.beta.fit(x))


@makedist("betabinom")
def MLbetabinomial(x):
    return sps.betabinom(*sps.betabinom.fit(x))


@makedist("binom")
def MLbinomial(x):
    """
    # TODO to check
    #! no fitting func for scipy discrete distributions
    """
    return sps.binom(*sps.binom.fit(x))


@makedist("chi2")
def MLchisquared(x):
    return sps.chi2(*sps.chi2.fit(x))


@makedist("expon")
def MLexponential(x):
    """a standalone caller for exponential distribution with interval data (not in use yet)"""
    if isinstance(x, sps.CensoredData | np.ndarray | list):
        return sps.expon(*sps.expon.fit(x))


@makedist("f")
def MLF(x):
    return sps.f(*sps.f.fit(x))


@makedist("gamma")
def MLgamma(x):
    return sps.gamma(*sps.gamma.fit(x))


@makedist("gammaexpon")
def MLgammaexponential(x):
    return sps.gammaexpon(*sps.gammaexpon.fit(x))


@makedist("geom")
def MLgeometric(x):
    return sps.geom(*sps.geom.fit(x))


@makedist("gumbel")
def MLgumbel(x):
    return sps.gumbel_r(*sps.gumbel_r.fit(x))


@makedist("laplace")
def MLlaplace(x):
    return sps.laplace(*sps.laplace.fit(x))


@makedist("logistic")
def MLlogistic(x):
    return sps.logistic(*sps.logistic.fit(x))


@makedist("lognorm")
def MLlognormal(x):
    return sps.lognorm(*sps.lognorm.fit(x))


# TODO why not use `sps.lognorm.fit(x)`` directly?
# def MLlognormal(x): return (sps.lognorm.rvs(
#     *sps.lognorm.fit(x), size=many))


@makedist("loguniform")
def MLloguniform(x):
    return sps.loguniform(*sps.loguniform.fit(x))


@makedist("nbinom")
def MLnegativebinomial(x):
    return sps.nbinom(*sps.nbinom.fit(x))


@makedist("norm")
def MLnormal(x):
    return sps.norm(*sps.norm.fit(x))


@makedist("pareto")
def MLpareto(x):
    return sps.pareto(*sps.pareto.fit(x))


@makedist("poisson")
def MLpoisson(x):
    return sps.poisson(*sps.poisson.fit(x))


@makedist("powerlaw")
def MLpowerfunction(x):
    return sps.powerlaw(*sps.powerlaw.fit(x))


@makedist("rayleigh")
def MLrayleigh(x):
    return sps.rayleigh(*sps.rayleigh.fit(x))


@makedist("t")
def MLstudent(x):
    return sps.t(*sps.t.fit(x))


@makedist("triang")
def MLtriangular(x):
    return sps.triang(*sps.triang.fit(x))


@makedist("uniform")
def MLuniform(x):
    return sps.uniform(*sps.uniform.fit(x))


named_mle_methods = {
    "bernoulli": MLbernoulli,
    "beta": MLbeta,
    "betabinomial": MLbetabinomial,
    "binomial": MLbinomial,
    "chisquared": MLchisquared,
    "exponential": MLexponential,
    "expon": MLexponential,
    "F": MLF,
    "f": MLF,
    "gamma": MLgamma,
    "gammaexponential": MLgammaexponential,
    "geometric": MLgeometric,
    "gumbel": MLgumbel,
    "laplace": MLlaplace,
    "logistic": MLlogistic,
    "lognormal": MLlognormal,
    "loguniform": MLloguniform,
    "negativebinomial": MLnegativebinomial,
    "norm": MLnormal,
    "normal": MLnormal,
    "gaussian": MLnormal,
    "pareto": MLpareto,
    "poisson": MLpoisson,
    "powerfunction": MLpowerfunction,
    "rayleigh": MLrayleigh,
    "student": MLstudent,
    "triangular": MLtriangular,
    "uniform": MLuniform,
}


##########################################################################
# Scott's maximum likelihood estimation constructors
##########################################################################

# Some of these functions may not support intervals in the data x           #**


def sMLbernoulli(x):
    return sps.bernoulli(x.mean())


def sMLnormal(x):
    return sps.norm(x.mean(), x.sd())  # **


def sMLgaussian(x):
    return MLnormal(x)


def sMLexponential(x):
    return sps.expon(x.mean())


def sMLpoisson(x):
    return sps.poisson(x.mean())


def sMLgeometric(x):
    return sps.geom(1 / (1 + x.mean()))


def sMLgumbel(x):
    loc, scale = sps.gumbel_r.fit(x)
    return sps.gumbel_r(loc, scale)


def sMLpascal(x):
    return MLgeometric(x)


def sMLuniform(x):
    return sps.uniform(min(x), max(x))  # **


def sMLrectangular(x):
    return MLuniform(x)


def sMLpareto(x):
    return sps.pareto(min(x), len(x) / np.sum(np.log(x) - np.log(min(x))))  # **


def sMLlaplace(x):
    return sps.laplace(x.median(), np.sum(np.abs(x - x.median()) / len(x)))  # **


def sMLdoubleexponential(x):
    return MLlaplace(x)


def sMLlognormal2(x):  # **
    n = len(x)
    mu = np.sum(np.log(x)) / n
    # this function gives clearly poor results
    return sps.lognormal2(mlog=mu, slog=np.sum((np.log(x) - mu) ** 2) / n)


# just uses transformation, which seems unlikely to be true, but fitdistrplus package uses it too
def sMLlognormal(x):
    return np.exp(MLnormal(np.log(x)))


def sMLloguniform(x):
    a, b, _, _ = sps.loguniform.fit(x)
    return sps.loguniform(a, b)


def sMLweibull(x, shapeinterval=None):  # **
    if shapeinterval is None:
        shapeinterval = np.array((0.001, 500))

    def wf(k):
        return (
            np.sum(x**k * np.log(x)) / np.sum(x**k) - np.sum(np.log(x)) / len(x) - 1 / k
        )

    k = sps.uniroot(wf, shapeinterval)
    el = np.exp(np.log(np.sum(x**k) / len(x)) / k)
    return sps.weibull_min.rvs(scale=el, shape=k)


def sMLgamma(data):  # **
    xbar = data.mean()
    shape = (xbar / data.sd()) ** 2  # initial estimate of shape from MoM
    logxbar = np.log(xbar)
    meanlog = np.log(data).mean()

    def f(x):
        return np.log(x) - np.digamma(x) - logxbar + meanlog

    shape = sps.uniroot(f, shape * np.array((0.5, 1.5))).root()
    rate = shape / xbar
    return sps.gamma(shape=shape, rate=rate)


named_smle_methods = {
    "bernoulli": sMLbernoulli,
    "normal": sMLnormal,
    "gaussian": sMLgaussian,
    "exponential": sMLexponential,
    "poisson": sMLpoisson,
    "geometric": sMLgeometric,
    "gumbel": sMLgumbel,
    "pascal": sMLpascal,
    "uniform": sMLuniform,
    "rectangular": sMLrectangular,
    "pareto": sMLpareto,
    "laplace": sMLlaplace,
    "doubleexponential": sMLdoubleexponential,
    "lognormal2": sMLlognormal2,
    "lognormal": sMLlognormal,
    "loguniform": sMLloguniform,
    "weibull": sMLweibull,
    "gamma": sMLgamma,
}


##########################################################################
# Maximum entropy distribution constructors
##########################################################################


def MEminmax(min, max):
    return sps.uniform(min, max)


# http://mathoverflow.net/questions/116667/whats-the-maximum-entropy-probability-distribution-given-bounds-a-b-and-mean, http://www.math.uconn.edu/~kconrad/blurbs/analysis/entropypost.pdf for discussion of this solution.
def MEminmaxmean(min, max, mean):
    return sps.sawinconrad(min, mean, max)


def MEmeansd(mean, sd):
    return sps.normal(mean, sd)


def MEminmean(min, mean):
    return min + sps.exponential(mean - min)


def MEdiscretemean(x, mu, steps=10, iterations=50):  # e.g., MEdiscretemean(1:10,2.3)
    x = np.array(x)

    def fixc(x, r):
        return 1 / np.sum(r**x)

    r = br = 1
    c = bc = fixc(x, r)
    d = bd = (mu - np.sum((c * r**x) * x)) ** 2
    for j in range(steps):
        step = 1 / (j + 1)
        for i in range(iterations):
            r = np.abs(br + (np.random.uniform() - 0.5) * step)
            c = fixc(x, r)
            d = (mu - np.sum((c * r**x) * x)) ** 2
            if d < bd:
                br = r
                bc = c
                bd = d
    w = bc * br**x
    w = w / np.sum(w)  # needed?

    z = np.array([])
    k = len(x)
    for i in range(k):
        z = np.concatenate((z, np.repeat(x[i], w[i] * many)))
    if len(z) >= many:
        z = z[0:many]
    else:
        z = np.concatenate((z, np.random.choice(x, size=many - len(z), p=w)))
    np.random.shuffle(z)  # shuffles z in place
    return z


def MEquantiles(v, p):
    if len(v) != len(p):
        stop("Inconsistent array lengths for quantiles")
    if (min(p) < 0) or (1 < max(p)):
        stop("Improper probability for quantiles")  # ensure 0 <= p <= 1
    if not (min(p) == 0 and max(p) == 1):
        stop("Probabilities must start at zero and go to one for quantiles")
    if np.any(np.diff(p) < 0):
        # ensure montone probabilities
        stop("Probabilities must increase for quantiles")
    if np.any(np.diff(v) < 0):
        stop("Quantiles values must increase")  # ensure montone quantiles
    x = np.repeat(np.inf, many)
    r = np.random.uniform(size=many)
    # np.where is Python's version of R's ifelse function
    for i in range(len(p) - 1):
        x = np.where(
            (p[i] <= r) & (r < p[i + 1]),
            v[i] + (r - p[i]) * (v[i + 1] - v[i]) / (p[i + 1] - p[i]),
            x,
        )
    return x


def MEdiscreteminmax(min, max):
    return np.minimum(np.trunc(uniform(min, max + 1)), max)


def MEmeanvar(mean, var):
    return MEmeansd(mean, np.sqrt(var))


def MEminmaxmeansd(min, max, mean, sd):
    return beta1((mean - min) / (max - min), sd / (max - min)) * (max - min) + min


def MEmmms(min, max, mean, sd):
    return beta1((mean - min) / (max - min), sd / (max - min)) * (max - min) + min


def MEminmaxmeanvar(min, max, mean, var):
    return MEminmaxmeansd(min, max, mean, np.sqrt(var))


###############################################################################
# Miscellaneous: PERT, Fermi methods, mean-normal range, KS, EDF, Antweiler
###############################################################################


# https://wernerantweiler.ca/blog.php?item=2019-06-05       #**
def antweiler(x):
    return sps.triang(min=min(x), mode=3 * np.mean(x) - max(x) - min(x), max=max(x))


def betapert(min, max, mode):  # N.B.  Not in numerical order!
    mu = (min + max + 4 * mode) / 6
    if abs(mode - mu) < 1e-8:
        alpha1 = alpha2 = 3
    else:
        alpha1 = (mu - min) * (2 * mode - min - max) / ((mode - mu) * (max - min))
        alpha2 = alpha1 * (max - mu) / (mu - min)
    return min + (max - min) * sps.beta(alpha1, alpha2)


def mnr(n, many=10000):
    xL = xU = np.random.normal(size=many)
    for i in range(n - 1):
        xx = np.random.normal(size=many)
        xL = np.minimum(xL, xx)
        xU = np.maximum(xU, xx)
    return np.mean(xU - xL)


def fermilnorm(x1, x2, n=None, pr=0.9):
    gm = np.sqrt(x1 * x2)
    if n is None:
        gsd = np.sqrt(x2 / x1) ** (1 / sps.norm.ppf(pr))  # qnorm(pr)
    else:
        gsd = np.exp((np.log(x2) - np.log(x1)) / mnr(n))
    return np.log((gm, gsd))


def ferminorm(x1, x2, n=None, pr=0.9):
    m = (x1 + x2) / 2
    if n is None:
        s = (x2 - x1) / (2 * sps.norm.ppf(pr))  # qnorm(pr)
    else:
        s = (x2 - x1) / mnr(n)
    return np.array((m, s))


def approxksD95(n):
    from scipy.interpolate import CubicSpline

    # approximations for the critical level for Kolmogorov-Smirnov statistic D,
    # for confidence level 0.95. Taken from Bickel & Doksum, table IX, p.483
    # and Lienert G.A.(1975) who attributes to Miller,L.H.(1956), JASA
    if n > 80:
        # Bickel&Doksum, table IX,p.483
        return 1.358 / (np.sqrt(n) + 0.12 + 0.11 / np.sqrt(n))
    else:
        x = np.array(
            (1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 15, 20, 30, 40, 50, 60, 70, 80)
        )  # from Lienert
        y = np.array(
            (
                0.975,
                0.84189,
                0.70760,
                0.62394,
                0.56328,  # 1:5
                0.51926,
                0.48342,
                0.45427,
                0.43001,
                0.40925,  # 6:10
                0.33760,
                0.29408,
                0.24170,
                0.21012,  # 15,20,30,40
                0.18841,
                0.17231,
                0.15975,
                0.14960,
            )
        )  # 50,60,70,80
        f = CubicSpline(x, y, bc_type="natural")
        return f(n)


def ks(x, conf=0.95, min=None, max=None):
    if conf != 0.95:
        stop("Cannot currently handle confidence levels other than 95%")
    h = histogram(x)
    mn = np.min(x)
    mx = np.max(x)
    if min is None:
        min = mn - (mx - mn) / 2
    if max is None:
        max = mx + (mx - mn) / 2
    lots = int(approxksD95(len(x)) * many)
    Lfermi = np.concatenate((lots * [min], h))
    Rfermi = np.concatenate((h, lots * [max]))
    Lfermi.sort()
    Rfermi[::-1].sort()
    # should prolly shuffle
    return np.concatenate((Lfermi[0:many], Rfermi[0:many]))


def ferminormconfband(x1, x2, n, pr=0.9, conf=0.95, bOt=0.001, tOp=0.999):
    if conf != 0.95:
        stop("Cannot handle confidence levels other than 95%")
    m, s = ferminorm(x1, x2, n, pr)
    lots = int(approxksD95(n) * many)
    BOT = scipy.stats.norm.ppf(bOt, m, s)
    TOP = scipy.stats.norm.ppf(tOp, m, s)
    Lfermi = np.concatenate((lots * [BOT], scipy.stats.norm.rvs(m, s, size=many)))
    Rfermi = np.concatenate((lots * [TOP], scipy.stats.norm.rvs(m, s, size=many)))
    Lfermi.sort()
    Rfermi[::-1].sort()
    # should prolly shuffle
    return np.concatenate((Lfermi[0:many], Rfermi[0:many]))


def fermilnormconfband(x1, x2, n, pr=0.9, conf=0.95, bOt=0.001, tOp=0.999):
    if conf != 0.95:
        stop("Cannot handle confidence levels other than 95%")
    mlog, slog = fermilnorm(x1, x2, n, pr)
    d = lognormal2(mlog, slog)
    lots = int(approxksD95(n) * many)
    BOT = scipy.stats.lognorm.ppf(bOt, s=slog, scale=np.exp(mlog))
    TOP = scipy.stats.lognorm.ppf(tOp, s=slog, scale=np.exp(mlog))
    Lfermi = np.concatenate((lots * [BOT], d))
    Rfermi = np.concatenate((lots * [TOP], d))
    Lfermi.sort()
    Rfermi[::-1].sort()
    # should prolly shuffle
    return np.concatenate((Lfermi[0:many], Rfermi[0:many]))


# * --------------- moments shape distribution constructors --------------- *#


# TODO: add more distributions in the future


def parse_moments(
    family: str,
    mean: Number | Interval,
    std: Number = None,
    var: Number = None,
    **kwargs,
):
    """parse the moments input to a standard form for dist construction

    args:
        family (str): distribution family
        mean (Number | Interval): mean value, which could be either precise value or an Interval object. Python list is not supported.
        std (Number): standard deviation
        var (Number): variance

    example:
        >>> parse_moments(family='normal', mean=3., std=1.)

    note:
        Only accept up to 2nd moment for now. Interval mean is supported due to single parameter construction.
        Combined mean and std intervals are not supported as they are deemed NP hard question.
    """

    # var --> std
    # Enforce that exactly one of std or var is given
    if std is not None and var is not None:
        raise ValueError("You cannot provide boh 'std' or 'var'.")

    # If variance is provided, convert to std
    if var is not None:
        try:
            std = np.sqrt(var)
        except:
            raise ValueError("Variance must be non-negative.")

    match family:
        # 1 parameter distributions
        case "exponential":
            return universal_exponential(mean=mean)
        case "rayleigh":
            return universal_rayleigh(mean=mean)
        # 2 parameter distributions
        case "normal" | "gaussian":
            return mm_normal(mean=mean, std=std)
        case "gamma":
            return mm_gamma(mean=mean, std=std)
        case "beta":
            return mm_beta(mean=mean, std=std)
        case "gumbel":
            return mm_gumbel(mean=mean, std=std)
        case "lognormal":
            return mm_lognormal(mean=mean, std=std)
        case _:
            raise ValueError(f"distribution family {family} not yet supported")



from __future__ import annotations
from typing import TYPE_CHECKING, Optional

from numpy.typing import NDArray
import pickle
from scipy.stats import wasserstein_distance, entropy
from sklearn.metrics import mean_squared_error
from dataclasses import dataclass
import matplotlib.pyplot as plt
import seaborn as sb
import pandas as pd
import time
import numpy as np
import logging
from ..console import console

if TYPE_CHECKING:
    from pyuncertainnumber import Distribution
    from .pdfs import ProbabilityDensityFun

"""
This is the implementation for Transitional Markov Chain Monte Carlo (TMCMC) algorithm

Leslie refactored, revised and integrated for `pyuncertainnumber` package, based on the version from 
Roberto Rocchetta (NASA UQ challenge 2025). The original code is from Mukesh K. Ramancha.
@license: MIT License
@date: Nov 2025
"""


logging.basicConfig(level=logging.INFO, format="%(asctime)s - %(message)s")


class TMCMC:
    """Class for TMCMC implementation

    args:
        N (int) : number of particles to be sampled from posterior

        parameters (list) : list of (size Nop) prior distributions instances

        log_likelihood (callable): log likelihood function on :math:`\\theta` to be defined which is problem specific

        status_file_name (str): name of the status file to store status of the tmcmc sampling

    return:
        mytrace: returns trace file of all samples of all tmcmc stages.
            At stage m: it contains [Sm, Lm, Wm_n, ESS, beta, Smcap]


    example:
        >>> from pyuncertainnumber.calibration.tmcmc import TMCMC
        >>> from pyuncertainnumber import pba
        >>> # create an instance of TMCMC class and run tmcmc
        >>> parameters = [pba.D("uniform", (0.8, 2.2)), pba.D("uniform", (0.4, 1.2))] # dummy prior distributions

        >>> t = TMCMC(
        >>>     N=1000,                                             # number of particles
        >>>     parameters=parameters,                              # prior distributions
        >>>     names =['k1', 'k2'],                                # parameter names
        >>>     log_likelihood=log_likelihood_function,             # user defined log likelihood function
        >>>     mutation_steps=1,                                   # number of MCMC steps for perturbation
        >>>     status_file_name='tmcmc_progressing_status.txt',    # status log file
        >>> )
        >>> mytrace = t.run()

        >>> # for loading trace file
        >>> import pickle
        >>> with open('tmcmc_trace.pkl', 'rb') as f:
        >>>     mytrace = pickle.load(f)
        >>> re = TMCMC(trace=mytrace, names=names)                  # create TMCMC instance with loaded trace and names

        >>> re.plot_updated_distribution(save=False)                # plot prior and posterior distribution
        >>> prior_samples = re.load_prior_samples()                 # load prior samples as DataFrame
        >>> posterior_samples = re.load_posterior_samples()         # load posterior samples as DataFrame


    .. figure:: /_static/tmcmc_2dof.png
        :alt: 2DOF TMCMC example
        :align: center
        :width: 70%

        Example of TMCMC calibration of a 2-DOF system
    """

    def __init__(
        self,
        N: int = None,
        parameters: list = None,
        names: list[str] = None,
        log_likelihood: callable = None,
        mutation_steps: int = 5,
        status_file_name: str = None,
        trace=None,
    ):
        self.N = N
        self.parameters = parameters
        self.names = names
        self.log_likelihood = log_likelihood
        self.mutation_steps = mutation_steps
        self.status_file_name = status_file_name
        self.trace = trace
        self.check_names()

    def check_names(self):
        if self.names is None:
            raise ValueError("Parameter names must be  provided.")

    def run(self) -> list[Stage]:
        """Run the TMCMC algorithm

        returns:
            mytrace: returns trace file of all samples of all tmcmc stages.
        """

        mytrace = run_tmcmc_updated(
            self.N,
            all_pars=self.parameters,
            log_likelihood=self.log_likelihood,
            Nm_steps_max=self.mutation_steps,
            status_file_name=self.status_file_name,
        )
        self.trace = mytrace
        self.check_trace()
        return mytrace

    def check_trace(self):
        if hasattr(self, "trace") and self.trace is not None:
            print("Trace is loaded.")

    def save_trace(self, mytrace, file_name: str, save_dir=None):
        """Save the trace to a file

        args:
            mytrace: trace file of all samples of all tmcmc stages.
            file_name (str): name of the file to save the trace
        """
        # save results
        if save_dir is not None:
            file_name = f"{save_dir}/{file_name}"

        with open(f"{file_name}.pkl", "wb") as f:
            pickle.dump(mytrace, f, protocol=pickle.HIGHEST_PROTOCOL)

    def load_prior_samples(self, trace=None) -> pd.DataFrame:
        if trace is None:
            trace = self.trace
        prior = pd.DataFrame(trace[0].samples, columns=self.names)  # samples from prior
        return prior

    def load_posterior_samples(self, trace=None) -> pd.DataFrame:
        if trace is None:
            trace = self.trace
        posterior = pd.DataFrame(trace[-1].samples, columns=self.names)
        return posterior

    def plot_updated_distribution(self, trace=None, save=False):
        """Plot the prior and posterior distribution of the parameters

        args:
            trace: trace file of all samples of all tmcmc stages.
            save (bool): if True, save the figure
        """
        if trace is None:
            trace = self.trace

        if not hasattr(self, "names") or self.names is None:
            raise ValueError("Parameter names are not provided.")

        plot_updated_distribution(trace, self.names, save=save)


@dataclass
class Stage:
    """Container for one TMCMC stage.

    Args:
        Sm (NDArray): samples before resampling (N x Np)
        Lm (NDArray): log-likelihoods at Sm (N,)
        Wm_n (NDArray): normalized incremental weights (N,)
        ESS (Optional[float]): effective sample size the next stage (can be None in terminal stage)
        beta (float): tempering parameter next stage beta
        Smcap (Optional[NDArray]): resampled samples (N x Np), None in terminal stage
    """

    Sm: NDArray  # samples before resampling (N x Np)
    Lm: NDArray  # log-likelihoods at Sm (N,)
    Wm_n: NDArray  # normalized incremental weights (N,)
    ESS: Optional[
        float
    ]  # effective sample size next stage ESS (can be None in terminal stage)
    beta: float  # tempering parameter next stage beta
    Smcap: Optional[NDArray]  # resampled samples (N x Np), None in terminal stage

    def __repr__(self) -> str:
        return (
            f"Stage(beta={self.beta:.4f}, " f"ESS={self.ESS}, " f"N={self.Sm.shape[0]})"
        )

    # * ---- Hints ----

    @property
    def samples(self) -> NDArray:
        """Samples before resampling (from previous stage)."""
        return self.Sm

    @property
    def log_likelihoods(self) -> NDArray:
        """Log-likelihood evaluated at `samples`."""
        return self.Lm

    @property
    def weights(self) -> NDArray:
        """Normalized incremental importance weights."""
        return self.Wm_n

    @property
    def effective_sample_size(self) -> Optional[float]:
        """Effective sample size after reweighting the next stage."""
        return self.ESS

    @property
    def tempering_parameter(self) -> float:
        """Tempering parameter beta next stage beta."""
        return self.beta

    @property
    def resampled_samples(self) -> Optional[NDArray]:
        """Samples after importance resampling (before MH mutation)."""
        return self.Smcap


def recover_trace_samples(mytrace: list[Stage], names: list[str]) -> pd.DataFrame:
    """Load back the Prior and Posterior samples in the form of a DataFrame

    args:
        mytrace (list[Stage]): trace file of all samples of all tmcmc stages.
        names (list[str]): list of parameter names

    returns:
        (pd.DataFrame) : Prior and Posterior in the trace as a DataFrame
    """

    df1 = pd.DataFrame(mytrace[0].samples, columns=names)  # samples from prior
    df2 = pd.DataFrame(
        mytrace[-1].samples, columns=names
    )  # samples from last step posterior

    # Add a column to identify the source
    df1["source"] = "Prior Samples"
    df2["source"] = "Posterior Samples"

    # Combine the DataFrames
    combined_df = pd.concat([df1, df2])

    # Combine the DataFrames
    combined_df = pd.concat([df1, df2])
    return combined_df


# * ----------------------------------- tmcmc


def plot_updated_distribution(mytrace, names: list[str], save=False):
    """Plot the prior and posterior distribution of the parameters

    args:
        mytrace: trace file of all samples of all tmcmc stages.
        names (list[str]): list of parameter names
        save (bool): if True, save the figure
    """

    # Example DataFrames from mytrace
    # names = list(Config.epistemic_domain.keys())

    # names = ["Xe_1", "Xe_2", "Xe_3", "Xa1_mu", "Xa2_mu", "Xa1_std", "Xa2_std"]
    df1 = pd.DataFrame(mytrace[0].samples, columns=names)  # samples from prior
    df2 = pd.DataFrame(
        mytrace[-1].samples, columns=names
    )  # samples from last step posterior

    # Add a column to identify the source
    df1["source"] = "Prior Samples"
    df2["source"] = "Posterior Samples"

    # Combine the DataFrames
    combined_df = pd.concat([df1, df2])

    # Combine the DataFrames
    combined_df = pd.concat([df1, df2])

    # Create the pairplot with density plots on the lower triangle
    g = sb.pairplot(
        combined_df,
        hue="source",
        palette="Set2",
        diag_kind="kde",  # Add KDE plots to the diagonal
    )

    # Add density plots on the lower triangle
    for i, j in zip(
        *np.tril_indices_from(g.axes, -1)
    ):  # Iterate over the lower triangle
        sb.kdeplot(
            data=combined_df,
            x=combined_df.columns[j],
            y=combined_df.columns[i],
            hue="source",
            fill=True,
            alpha=0.5,
            ax=g.axes[i, j],
        )
    if save == True:
        print("Note: figure saved")
        plt.savefig("updated_distribution.png")
    plt.show()


def initial_population(N: float, all_pars: list) -> NDArray:
    """Generates initial population from prior distribution

    Args:
        N (float): number of particles.

        all_pars (list) : All the parameters, of size Np, to be updated. `all_pars[i]` is object of type `pdfs`.

    Returns
        ini_pop (NDArray): the initial population which is a numpy array of size `N x Np`.

    note:
        `Np` is number of parameters
    """
    ini_pop = np.zeros((N, len(all_pars)))
    for i in range(len(all_pars)):
        ini_pop[:, i] = all_pars[i].generate_rns(N)
    return ini_pop


def log_prior(
    s: NDArray, all_pars: list[ProbabilityDensityFun | Distribution]
) -> float:
    """Computes log_prior value at all particles

    Args:
        s (NDArray): numpy array of size (N x Np) of all particles.

        all_pars (list[ProbabilityDensityFun | Distribution]) : All the parameters, of size Np, to be updated. `all_pars[i]` is object of type `pdfs`.

    Returns
        log_p (NDArray): log prior at all N particles which is a numpy array of size N

    Note:
        `Np` denotes number of parameters
    """
    log_p = 0
    for i in range(len(s)):
        log_p = log_p + all_pars[i].log_pdf_eval(s[i])
    return log_p


def compute_beta_update_evidence(
    beta: float, log_likelihoods: NDArray, log_evidence: float, prev_ESS: int
):
    """Computes beta for the next stage and updated model evidence

    Args:
        beta (float): stage parameter.
        log_likelihoods (NDArray): log likelihood values at all particles which is numpy array of size N
        log_evidence (float): log of evidence.
        prev_ESS (int): effective sample size of previous stage

    :returns: A 4-tuple ``(new_beta, log_evidence, Wm_n, ESS)`` where:
        - **new_beta** (*float*): stage parameter for next stage.
        - **log_evidence** (*float*): updated log evidence.
        - **Wm_n** (*NDArray*): weights of particles for the next stage.
        - **ESS** (*float*): effective sample size of new stage.
    :rtype: tuple[float, float, NDArray, float]

    """
    old_beta = beta
    min_beta = beta
    max_beta = 2.0
    N = len(log_likelihoods)
    # rN = int(len(log_likelihoods) * 0.5)
    # rN = 0.95*prev_ESS
    rN = max(0.95 * prev_ESS, 50)  # min particles 50

    while max_beta - min_beta > 1e-8:  # min step size
        new_beta = 0.5 * (max_beta + min_beta)

        # plausible weights of Sm corresponding to new beta
        inc_beta = new_beta - old_beta

        Wm = np.exp(inc_beta * (log_likelihoods - log_likelihoods.max()))
        Wm_n = Wm / sum(Wm)
        ESS = int(1 / np.sum(Wm_n**2))

        # log_Wm = inc_beta * log_likelihoods
        # log_Wm_n = log_Wm - logsumexp(log_Wm)
        # ESS = int(np.exp(-logsumexp(log_Wm_n * 2)))

        if ESS == rN:
            break
        elif ESS < rN:
            max_beta = new_beta
        else:
            min_beta = new_beta

    if new_beta >= 1:
        new_beta = 1

        # plausible weights of Sm corresponding to new beta
        inc_beta = new_beta - old_beta

        Wm = np.exp(inc_beta * (log_likelihoods - log_likelihoods.max()))
        Wm_n = Wm / sum(Wm)

        # log_Wm = inc_beta * log_likelihoods
        # log_Wm_n = log_Wm - logsumexp(log_Wm)

    # Wm = np.exp(log_Wm)
    # Wm_n = np.exp(log_Wm_n)

    # update model evidence
    # (check it, might not be correct, as we remove log.likelihood max in compute_beta)
    # evidence = evidence * (sum(Wm)/N)
    # log_evidence = log_evidence + logsumexp(log_Wm) - np.log(N)
    log_evidence = log_evidence + np.log((sum(Wm) / N))

    return new_beta, log_evidence, Wm_n, ESS


def propose(current: NDArray, covariance: NDArray, n: int) -> NDArray:
    """Proposal distribution for MCMC in pertubation stage

    Args:
        current (NDArray): numpy array of size Np
            current particle location
        covariance (NDArray): numpy array of size Np x Np
            proposal covariance matrix
        n (int): number of proposals.

    Returns:
        NDArray: n proposals
    """
    return np.random.multivariate_normal(current, covariance, n)


def MCMC_MH(
    particle_num: int,
    Em: NDArray,
    Nm_steps: int,
    current: NDArray,
    likelihood_current: float,
    posterior_current: float,
    beta: float,
    numAccepts: int,
    all_pars: list,
    log_likelihood: callable,
) -> tuple[NDArray, float, float, int]:
    """Markov chain Monte Carlo using Metropolis-Hastings which "perturbs" each particle using MCMC-MH

    Conduct `Nm_steps` steps of MH for each particle.

    Args:
        particle_num (int) : The index of the current particle/parameter vector being evaluated.

        Em (NDArray) : proposal covarince matrix which is numpy array of size Nop x Nop

        Nm_steps (int) : number of perturbation steps.

        current (NDArray) : numpy array of size Nop. current particle location

        likelihood_current (float) : log likelihood value at current particle

        posterior_current (float) : log posterior value at current particle

        beta (float) : stage parameter.

        numAccepts (int) : total number of accepts

        all_pars : list of size Nop
            Nop is number of parameters
            all_pars[i] is object of type pdfs
            all parameters to be inferred.

        log_likelihood (callable): log likelihood function to be defined in main.py.

    :returns: A 4-tuple ``(current, likelihood_current, posterior_current, numAccepts)`` where:

        - **current** (*NDArray*): Perturbed particle location of shape (Nop,).
        - **likelihood_current** (*float*): Log-likelihood at the perturbed particle.
        - **posterior_current** (*float*): Log-posterior at the perturbed particle.
        - **numAccepts** (*int*): Updated total number of accepts.
    :rtype: tuple[NDArray, float, float, int]

    """
    all_proposals = []
    all_PLP = []

    deltas = propose(np.zeros(len(current)), Em, Nm_steps)
    # logging.info(
    #     f"MCMC_MH: running for particle_num = {particle_num}, .... N chains {Nm_steps}"
    # )

    for j2 in range(Nm_steps):
        delta = deltas[j2]
        proposal = current + delta
        prior_proposal = log_prior(proposal, all_pars)

        if not np.isfinite(prior_proposal):

            likelihood_proposal = -np.inf  # dont run the model
            posterior_proposal = -np.inf
        # proposal satisfies the prior constraints
        else:
            likelihood_proposal = log_likelihood(particle_num, proposal)
            posterior_proposal = prior_proposal + likelihood_proposal * beta

        log_acceptance = posterior_proposal - posterior_current

        all_proposals.append(proposal)
        all_PLP.append([prior_proposal, likelihood_proposal, posterior_proposal])

        if np.isfinite(log_acceptance) and (
            np.log(np.random.uniform()) < log_acceptance
        ):
            # accept
            current = proposal
            posterior_current = posterior_proposal
            likelihood_current = likelihood_proposal
            numAccepts += 1

    # gather all last samples
    return current, likelihood_current, posterior_current, numAccepts


# TODO
def run_tmcmc_vec():
    """vectorised version of run_tmcmc to be implemented later"""
    pass


def run_tmcmc_updated(
    N: int,
    all_pars: list,
    log_likelihood: callable,
    status_file_name: str,
    Nm_steps_max: int = 5,
    Nm_steps_maxmax: int = 5,
) -> list[Stage]:
    """Updated main workflow of running Transitional MCMC

    Solves the "At this point, beta has been updated to the new value, but Postm is still the log-target from the previous beta (except at the very first stage).
    So Postmcap is inconsistent with the beta you pass into MCMC_MH."

    args:
        N  (int) : int
            number of particles to be sampled from posterior

        all_pars (list) : All the parameters, of size Nop, to be updated. `all_pars[i]` is object of type `pdfs`.

        log_likelihood (callable): log likelihood function to be defined in main.py as is problem specific

        status_file_name (str): name of the status file to store status of the tmcmc sampling

        Nm_steps_max (int, optional): Numbers of MCMC steps for perturbation. The default is 5.

        Nm_steps_maxmax (int, optional): Numbers of MCMC steps for perturbation. The default is 5.

    returns:
        list[Stage]: the trace that contains all iterations of the Stage instances.
    """
    parallel_processing = "multiprocessing"

    # side note: make all_pars as ordered dict in the future
    # Initialize (beta, effective sample size)
    beta = 0
    ESS = N
    mytrace = []
    stage_num = 0
    start_time_global = time.time()

    # Initialize other TMCMC variables
    Nm_steps = Nm_steps_max
    parallelize_MCMC = True
    Adap_calc_Nsteps, Adap_scale_cov = "yes", "yes"  # yes or no
    scalem = 1  # cov scale factor
    log_evidence = 0  # model evidence

    # initial samples -> array (N, Np)
    Sm = initial_population(N, all_pars)

    # Evaluate posterior at Sm
    Priorm = np.array([log_prior(s, all_pars) for s in Sm]).squeeze()
    Postm = Priorm  # prior = post for beta = 0

    status_file = open(status_file_name, "a+")

    # Evaluate log-likelihood at current samples Sm
    if parallelize_MCMC:
        iterables = [(ind, Sm[ind]) for ind in range(N)]
        status_file.write("======================== \n")
        if parallel_processing == "multiprocessing":
            status_file.write("using multiprocessing \n")
            import multiprocessing as mp
            from multiprocessing import Pool

            pool = Pool(processes=mp.cpu_count() - 2)
            Lmt = pool.starmap(log_likelihood, iterables)

        else:
            raise (
                AssertionError(
                    "parallel_processing invalid, should be either multiprocessing or mpi"
                )
            )
        status_file.write("======================== \n")
        Lm = np.array(Lmt).squeeze()
    else:
        Lm = []
        for ind in range(N):
            Lm.append(log_likelihood(ind, Sm[ind]))
            logging.info(
                f"Computing likelihood stage {stage_num}: particle:{ind + 1}/{N}"
            )
        Lm = np.array(Lm).squeeze()
    status_file.close()

    while beta < 1:
        stage_num += 1
        start_time_stage = time.time()

        # adaptivly compute beta s.t. ESS = N/2 or ESS = 0.95*prev_ESS
        # plausible weights of Sm corresponding to new beta
        # logging.info(f"'Computing beta the weights ...")

        beta_old = beta
        beta, log_evidence, Wm_n, ESS = compute_beta_update_evidence(
            beta, Lm, log_evidence, ESS
        )

        # NEW: update Postm to match the NEW beta, before resampling / MH
        Postm = Postm + (beta - beta_old) * Lm

        console.log(
            f"[bold green]TMCMC Iteration stage {stage_num}: Tempering parameter updated to {beta:.6f}[/bold green]"
        )

        # Calculate covaraince matrix using Wm_n
        Cm = np.cov(Sm, aweights=Wm_n, rowvar=0)

        # * --------------------------------------------------------- Resample
        # Resampling using plausible weights
        SmcapIDs = np.random.choice(range(N), N, p=Wm_n)
        # SmcapIDs = resampling.stratified_resample(Wm_n)
        Smcap = Sm[SmcapIDs]  # Resampled particles
        Lmcap = Lm[SmcapIDs]
        Postmcap = Postm[SmcapIDs]

        #  save to trace
        # stage m: samples, likelihood, weights, next stage ESS, next stage beta, resampled samples
        mytrace.append(Stage(Sm=Sm, Lm=Lm, Wm_n=Wm_n, ESS=ESS, beta=beta, Smcap=Smcap))

        # TODO: plot updated distribution to fix later on
        # if stage_num in [2, 4, 6]:
        #     plot_updated_distribution(mytrace)

        # print to status_file
        status_file = open(status_file_name, "a+")
        status_file.write("stage number = %d \n" % stage_num)
        status_file.write("beta = %.5f \n" % beta)
        status_file.write("ESS = %d \n" % ESS)
        status_file.write("scalem = %.2f \n" % scalem)

        # * --------------------------------------------------------- MH Perturb
        # perform MCMC starting at each Smcap (total: N) for Nm_steps
        Em = ((scalem) ** 2) * Cm  # Proposal dist covariance matrix

        numProposals = N * Nm_steps
        numAccepts = 0

        if parallelize_MCMC:
            iterables = [
                (
                    j1,
                    Em,
                    Nm_steps,
                    Smcap[j1],
                    Lmcap[j1],
                    Postmcap[j1],
                    beta,
                    numAccepts,
                    all_pars,
                    log_likelihood,
                )
                for j1 in range(N)
            ]

            if parallel_processing == "multiprocessing":
                results = pool.starmap(MCMC_MH, iterables)

            # elif parallel_processing == "mpi":
            #     results = list(executor.starmap(MCMC_MH, iterables))
        else:
            """Here we are running Markov-Chain Monte Carlo for each particle"""
            results = [
                MCMC_MH(
                    j1,
                    Em,
                    Nm_steps,
                    Smcap[j1],
                    Lmcap[j1],
                    Postmcap[j1],
                    beta,
                    numAccepts,
                    all_pars,
                    log_likelihood,
                )
                for j1 in range(N)
            ]

        Sm1, Lm1, Postm1, numAcceptsS = zip(*results)
        Sm1 = np.asarray(Sm1)
        Lm1 = np.asarray(Lm1)
        Postm1 = np.asarray(Postm1)
        numAcceptsS = np.asarray(numAcceptsS)
        numAccepts = sum(numAcceptsS)

        # total observed acceptance rate
        R = numAccepts / numProposals
        status_file.write("acceptance rate = %.2f \n" % R)

        # Calculate Nm_steps based on observed acceptance rate
        if Adap_calc_Nsteps == "yes":
            # increase max Nmcmc with stage number
            Nm_steps_max = min(Nm_steps_max + 1, Nm_steps_maxmax)
            status_file.write("adapted max MCMC steps = %d \n" % Nm_steps_max)

            acc_rate = max(1.0 / numProposals, R)
            Nm_steps = min(
                Nm_steps_max, 1 + int(np.log(1 - 0.99) / np.log(1 - acc_rate))
            )
            status_file.write("next MCMC Nsteps = %d \n" % Nm_steps)

        status_file.write("log_evidence till now = %.20f \n" % log_evidence)
        status_file.write(
            "--- Execution time: %.2f mins --- \n"
            % ((time.time() - start_time_stage) / 60)
        )
        status_file.write("======================== \n")
        status_file.close()

        # scale factor based on observed acceptance ratio
        if Adap_scale_cov == "yes":
            scalem = (1 / 9) + ((8 / 9) * R)

        # for next beta
        Sm, Postm, Lm = Sm1, Postm1, Lm1

    # save to trace, which is the last stopping stage
    mytrace.append(
        Stage(
            Sm=Sm,
            Lm=Lm,
            Wm_n=np.ones(len(Wm_n)) / len(Wm_n),
            ESS=None,
            beta=1.0,
            Smcap=None,
        )
    )

    status_file = open(status_file_name, "a+")
    status_file.write(
        "--- Execution time: %.2f mins --- \n"
        % ((time.time() - start_time_global) / 60)
    )
    status_file.write("log_evidence = %.20f \n" % log_evidence)

    if parallelize_MCMC:
        if parallel_processing == "multiprocessing":
            status_file.write("closing multiprocessing \n")
            pool.close()

    status_file.close()

    if parallel_processing == "multiprocessing":
        return mytrace


def run_tmcmc(
    N: int,
    all_pars: list,
    log_likelihood: callable,
    status_file_name: str,
    Nm_steps_max: int = 5,
    Nm_steps_maxmax: int = 5,
):
    """Main workflow of running Transitional MCMC

    args:
        N  (int) : int
            number of particles to be sampled from posterior

        all_pars (list) : All the parameters, of size Nop, to be updated. `all_pars[i]` is object of type `pdfs`.

        log_likelihood (callable): log likelihood function to be defined in main.py as is problem specific

        status_file_name (str): name of the status file to store status of the tmcmc sampling

        Nm_steps_max (int, optional): Numbers of MCMC steps for perturbation. The default is 5.

        Nm_steps_maxmax (int, optional): Numbers of MCMC steps for perturbation. The default is 5.

        parallel_processing (str): should be either 'multiprocessing' or 'mpi'

    returns:
        a tuple containing:
            mytrace: returns trace file of all samples of all tmcmc stages.
                at stage m: it contains [Sm, Lm, Wm_n, ESS, beta, Smcap]
    """
    parallel_processing = "multiprocessing"

    # side note: make all_pars as ordered dict in the future
    # Initialize (beta, effective sample size)
    beta = 0
    ESS = N
    mytrace = []
    stage_num = 0
    start_time_global = time.time()

    # Initialize other TMCMC variables
    Nm_steps = Nm_steps_max
    parallelize_MCMC = True
    Adap_calc_Nsteps, Adap_scale_cov = "yes", "yes"  # yes or no
    scalem = 1  # cov scale factor
    log_evidence = 0  # model evidence

    # initial samples -> array (N, Np)
    Sm = initial_population(N, all_pars)

    # Evaluate posterior at Sm
    Priorm = np.array([log_prior(s, all_pars) for s in Sm]).squeeze()
    Postm = Priorm  # prior = post for beta = 0

    status_file = open(status_file_name, "a+")

    # Evaluate log-likelihood at current samples Sm
    if parallelize_MCMC:
        iterables = [(ind, Sm[ind]) for ind in range(N)]
        status_file.write("======================== \n")
        if parallel_processing == "multiprocessing":
            status_file.write("using multiprocessing \n")
            import multiprocessing as mp
            from multiprocessing import Pool

            pool = Pool(processes=mp.cpu_count() - 2)
            Lmt = pool.starmap(log_likelihood, iterables)

        else:
            raise (
                AssertionError(
                    "parallel_processing invalid, should be either multiprocessing or mpi"
                )
            )
        status_file.write("======================== \n")
        Lm = np.array(Lmt).squeeze()
    else:
        Lm = []
        for ind in range(N):
            Lm.append(log_likelihood(ind, Sm[ind]))
            logging.info(
                f"Computing likelihood stage {stage_num}: particle:{ind + 1}/{N}"
            )
        Lm = np.array(Lm).squeeze()
    status_file.close()

    while beta < 1:
        stage_num += 1
        start_time_stage = time.time()

        # adaptivly compute beta s.t. ESS = N/2 or ESS = 0.95*prev_ESS
        # plausible weights of Sm corresponding to new beta
        # logging.info(f"'Computing beta the weights ...")

        beta, log_evidence, Wm_n, ESS = compute_beta_update_evidence(
            beta, Lm, log_evidence, ESS
        )

        console.log(
            f"[bold green]TMCMC Iteration stage {stage_num}: Tempering parameter updated to {beta:.6f}[/bold green]"
        )

        # Calculate covaraince matrix using Wm_n
        Cm = np.cov(Sm, aweights=Wm_n, rowvar=0)

        # * --------------------------------------------------------- Resample
        # Resampling using plausible weights
        SmcapIDs = np.random.choice(range(N), N, p=Wm_n)
        # SmcapIDs = resampling.stratified_resample(Wm_n)
        Smcap = Sm[SmcapIDs]  # Resampled particles
        Lmcap = Lm[SmcapIDs]
        Postmcap = Postm[SmcapIDs]

        #  save to trace
        # stage m: samples, likelihood, weights, next stage ESS, next stage beta, resampled samples
        mytrace.append([Sm, Lm, Wm_n, ESS, beta, Smcap])

        # TODO: plot updated distribution to fix later on
        # if stage_num in [2, 4, 6]:
        #     plot_updated_distribution(mytrace)

        # print to status_file
        status_file = open(status_file_name, "a+")
        status_file.write("stage number = %d \n" % stage_num)
        status_file.write("beta = %.5f \n" % beta)
        status_file.write("ESS = %d \n" % ESS)
        status_file.write("scalem = %.2f \n" % scalem)

        # * --------------------------------------------------------- MH Perturb
        # perform MCMC starting at each Smcap (total: N) for Nm_steps
        Em = ((scalem) ** 2) * Cm  # Proposal dist covariance matrix

        numProposals = N * Nm_steps
        numAccepts = 0

        if parallelize_MCMC:
            iterables = [
                (
                    j1,
                    Em,
                    Nm_steps,
                    Smcap[j1],
                    Lmcap[j1],
                    Postmcap[j1],
                    beta,
                    numAccepts,
                    all_pars,
                    log_likelihood,
                )
                for j1 in range(N)
            ]

            if parallel_processing == "multiprocessing":
                results = pool.starmap(MCMC_MH, iterables)

            # elif parallel_processing == "mpi":
            #     results = list(executor.starmap(MCMC_MH, iterables))
        else:
            """Here we are running Markov-Chain Monte Carlo for each particle"""
            results = [
                MCMC_MH(
                    j1,
                    Em,
                    Nm_steps,
                    Smcap[j1],
                    Lmcap[j1],
                    Postmcap[j1],
                    beta,
                    numAccepts,
                    all_pars,
                    log_likelihood,
                )
                for j1 in range(N)
            ]

        Sm1, Lm1, Postm1, numAcceptsS = zip(*results)
        Sm1 = np.asarray(Sm1)
        Lm1 = np.asarray(Lm1)
        Postm1 = np.asarray(Postm1)
        numAcceptsS = np.asarray(numAcceptsS)
        numAccepts = sum(numAcceptsS)

        # total observed acceptance rate
        R = numAccepts / numProposals
        status_file.write("acceptance rate = %.2f \n" % R)

        # Calculate Nm_steps based on observed acceptance rate
        if Adap_calc_Nsteps == "yes":
            # increase max Nmcmc with stage number
            Nm_steps_max = min(Nm_steps_max + 1, Nm_steps_maxmax)
            status_file.write("adapted max MCMC steps = %d \n" % Nm_steps_max)

            acc_rate = max(1.0 / numProposals, R)
            Nm_steps = min(
                Nm_steps_max, 1 + int(np.log(1 - 0.99) / np.log(1 - acc_rate))
            )
            status_file.write("next MCMC Nsteps = %d \n" % Nm_steps)

        status_file.write("log_evidence till now = %.20f \n" % log_evidence)
        status_file.write(
            "--- Execution time: %.2f mins --- \n"
            % ((time.time() - start_time_stage) / 60)
        )
        status_file.write("======================== \n")
        status_file.close()

        # scale factor based on observed acceptance ratio
        if Adap_scale_cov == "yes":
            scalem = (1 / 9) + ((8 / 9) * R)

        # for next beta
        Sm, Postm, Lm = Sm1, Postm1, Lm1

    # save to trace, which is the last stopping stage
    mytrace.append([Sm, Lm, np.ones(len(Wm_n)) / len(Wm_n), "notValid", 1, "notValid"])

    status_file = open(status_file_name, "a+")
    status_file.write(
        "--- Execution time: %.2f mins --- \n"
        % ((time.time() - start_time_global) / 60)
    )
    status_file.write("log_evidence = %.20f \n" % log_evidence)

    if parallelize_MCMC:
        if parallel_processing == "multiprocessing":
            status_file.write("closing multiprocessing \n")
            pool.close()

    status_file.close()

    if parallel_processing == "multiprocessing":
        return mytrace


# * ----------------------------------- likelihood


def log_likelihood(
    sam_id: int,
    param_vec: NDArray,
    data_empirical,
    n_xa_samples,
    simulator: callable,
    which_likelihood_calculator="gaussian",
) -> np.ndarray:
    """Picklable function at module level.

    args:
        sam_id: (int) a dummy sample index  which is required in the main `run_tmcmc` workflow

        param_vec (array-like): indeed a 1d epistemic_param_vec, the parameter vector, i.e. epistemic_domain


    note:
        Gaussian approximate likelihood function is implemented here. Other likelihood functions such as
        "pseudo" and "vae" are underway. `sam_id` is not used here, but it is required in the main `run_tmcmc` workflow
    """

    if which_likelihood_calculator == "gaussian":
        llhd_calculator = gaussian_likelihood_fun
    elif which_likelihood_calculator == "pseudo":
        llhd_calculator = "pseudo"
    elif which_likelihood_calculator == "vae":
        llhd_calculator = "vae"
    else:
        raise ValueError(f"Unknown likelihood function: {which_likelihood_calculator}")

    pseudo_log_likelihoods = llhd_calculator(
        params=param_vec,
        data=data_empirical,
        simulator=simulator,
        n_xa_samples=n_xa_samples,
    )
    return np.sum(pseudo_log_likelihoods, axis=-1)


def gaussian_likelihood_fun(
    params, data, simulator, n_xa_samples, dissimilarity_metric="wasserstein"
):
    """Computes the Gaussian log-likelihood of the data given the model and parameters.

    args:
        params: Parameters for the black-box model.

        data: Empirical data (numpy array of shape (time_steps, features, samples)).

        simulator: A function that generates model response samples given parameters.
            Here response may be time series data of shape (time_steps, features, n_xa_samples), or
            maybe some performance metrics extracted from the time series data.

        n_xa_samples: Number of samples to generate from the simulator.

    returns:
        log_likelihood: The Gaussian log-likelihood of the data given the model.
    """

    # Generate samples from the black box model using params
    mod_sam = simulator(params, n_xa_samples)

    # Compute global dissimilarity
    if dissimilarity_metric == "mse":
        # Mean Squared Error
        dissimilarity = mean_squared_error(data.flatten(), mod_sam.flatten())

        # time series case
        # data_expanded = np.expand_dims(data, axis=-1)  # Shape: (60, 6, 100, 1)
        # mod_sam_expanded = np.expand_dims(mod_sam, axis=-2)  # Shape: (60, 6, 1, 10)
        # dissimilarity = np.mean((data_expanded - mod_sam_expanded) ** 2)

    elif dissimilarity_metric == "wasserstein":
        dissimilarity = wasserstein_distance(data.flatten(), mod_sam.flatten())

    elif dissimilarity_metric == "correlation":
        corr = np.corrcoef(data.flatten(), mod_sam.flatten())[0, 1]
        dissimilarity = 1 - corr  # Correlation-based distance

    else:
        raise ValueError(f"Unknown dissimilarity metric: {dissimilarity_metric}")

    # Convert dissimilarity to likelihood
    eps_scale = 1e-3  # scaling factor for likelihood
    log_likelihood = -((dissimilarity / eps_scale) ** 2)  # np.exp(-dissimilarity)

    return log_likelihood


# * ----------------------------------- helper functions


def get_top_samples(trace, top_num=20):
    """Get the top posterior samples based on likelihood values."""
    posterior_samples = trace[-1].samples  # shape (N, 8)
    likelihoods = trace[-1].log_likelihoods  # shape (N,)

    # Select top 20% samples by likelihood
    top_indices = np.argsort(likelihoods)[-top_num:]
    top_samples = posterior_samples[top_indices]
    return top_samples


def get_posterior_samples(trace: list[Stage]) -> NDArray:
    """Get the posterior particles of the last stage from the trace.

    args:
        trace (list[Stage]): trace object.
    """
    posterior_samples = trace[-1].samples  # shape (N, 8)
    return posterior_samples


def hdi_1d(x: NDArray, cred_mass=0.95) -> tuple[float, float]:
    """Obtain the highest density interval (HDI) from particles

    args:
        x (NDArray): 1D array of samples.
        cred_mass (float): Credible mass for the HDI.

    returns:
        tuple: Lower and upper bounds of the HDI.
    """
    x = np.asarray(x, dtype=float)
    x = x[np.isfinite(x)]
    if x.size == 0:
        return (np.nan, np.nan)

    x = np.sort(x)
    n = x.size
    m = int(np.floor(cred_mass * n))
    if m < 1:
        return (np.nan, np.nan)

    low = x[: n - m]
    high = x[m:]
    widths = high - low
    j = np.argmin(widths)

    return (low[j], high[j])


def get_hdi_bounds(
    data: pd.DataFrame | NDArray, levels=(0.99, 0.95, 0.20, 0.10, 0.05)
) -> pd.DataFrame:
    """Compute HDI bounds for each column in `data`

    args:
        data (pd.DataFrame | NDArray): posterior samples (shape: n_samples x n_vars).
        levels (tuple): A tuple of high density intervalc levels.

    returns:
        pd.DataFrame: DataFrame with HDI bounds for each column and level.
    """
    # Normalize input
    if isinstance(data, pd.DataFrame):
        values = data.to_numpy()
        columns = data.columns
    else:
        values = np.asarray(data)
        if values.ndim != 2:
            raise ValueError("Input array must be 2D (n_samples, n_columns)")
        columns = [f"col_{i}" for i in range(values.shape[1])]

    out = {}

    for j, col in enumerate(columns):
        col_vals = values[:, j]
        for lvl in levels:
            lo, hi = hdi_1d(col_vals, cred_mass=lvl)
            p = int(round(lvl * 100))
            out.setdefault(f"hdi_{p}_low", {})[col] = lo
            out.setdefault(f"hdi_{p}_high", {})[col] = hi

    return pd.DataFrame(out).rename_axis("column").reset_index()


def transform_old_trace_to_new(trace: list[list]) -> list[Stage]:
    """Transform old trace format to new Stage class format.

    args:
        trace (list[list]): Old trace instance which is list of lists, where each inner list contains:
            [Sm, Lm, Wm_n, ESS, beta, Smcap]

    returns:
        list[Stage]: New trace instance which is list of Stage class instances.
    """
    return [Stage(*i) for i in trace]



import numpy as np
import scipy.stats as sps
from pyuncertainnumber import pba
import pyuncertainnumber as pun


def test_mom():
    """test the MOM fitting method"""
    precise_sample = sps.expon(scale=1 / 0.4).rvs(15)
    imprecise_data = pba.I(lo=precise_sample - 1.4, hi=precise_sample + 1.4)

    # parametric distributional estimator using method of matching moments
    expon_fitted = pun.fit("mom", family="exponential", data=imprecise_data)
    assert isinstance(expon_fitted, pun.UncertainNumber)


def test_mle():
    """test the MLE fitting method"""
    u = pun.fit("mle", "norm", np.random.normal(0, 1, 100))
    assert isinstance(u, pun.UncertainNumber)



import pytest
import pyuncertainnumber as pun
from pyuncertainnumber import pba
from pyuncertainnumber.propagation.performance import foo_universal
import numpy as np
import pyuncertainnumber.propagation.mixed_up as mix


@pytest.fixture
def example_pboxes():
    a = pba.normal([2, 3], [1])
    b = pba.normal([10, 14], [1])
    c = pba.uniform([4, 5], [10, 11])
    return a, b, c


# * ------------------------- pba arithmetic


def test_naive_arithmetic(example_pboxes):
    a, b, c = example_pboxes

    with pba.dependency("i"):
        t = foo_universal([a, b, c])


# * ------------------------- low level


def test_low_level_imc(example_pboxes):
    """test Interval Monte Carlo method"""
    a, b, c = example_pboxes

    corre_matrix = np.array([[1, 0.5, 0.3], [0.5, 1, 0.4], [0.3, 0.4, 1]])
    de = pba.Dependency(family="gaussian", corr=corre_matrix)
    t_gau_copula = mix.interval_monte_carlo(
        vars=[a, b, c],
        func=foo_universal,
        dependency=de,
        interval_strategy="direct",
        n_sam=10,
    )  # `direct` strategy succeeds
    assert isinstance(t_gau_copula, pun.pba.Pbox)

    t1_independence = mix.interval_monte_carlo(
        vars=[a, b, c], func=foo_universal, interval_strategy="direct", n_sam=10
    )  # `direct` strategy succeeds

    assert isinstance(t1_independence, pun.pba.Pbox)

    t1 = mix.interval_monte_carlo(
        vars=[a, b, c],
        func=foo_universal,
        interval_strategy="endpoints",
        n_sam=10,
    )  # `endpoints` strategy succeeds (expects foo_vec)
    assert isinstance(t1, pun.pba.Pbox)

    t1_sub = mix.interval_monte_carlo(
        vars=[a, b, c],
        func=foo_universal,
        interval_strategy="subinterval",
        n_sam=10,
        n_sub=10,
        subinterval_style="endpoints",
    )  # `subinterval` strategy succeeds

    assert isinstance(t1_sub, pun.pba.Pbox)


def test_low_level_slicing(example_pboxes):
    a, b, c = example_pboxes
    t3 = mix.slicing(
        vars=[a, b, c], func=foo_universal, interval_strategy="endpoints", n_slices=10
    )  # `endpoints` strategy succeeds

    assert isinstance(t3, pun.pba.Pbox)



from pathlib import Path
import csv
import pandas as pd
import numpy as np

# TODO Not quite sure if data saving behavior will be consistent across all aleatory, epistemic, mixed methods
# TODO what will be saved for aleatory case? what will be saved for mixed case?


class Propagation_results:
    """Stores the results of uncertainty propagation with multiple outputs, sharing raw_data x and f.

    args:
        un (np.ndarray): np.array of UncertainNumber objects (one for each output).
        raw_data (dict): Dictionary containing raw data shared across outputs:
            x (np.ndarray): Input values.
            f (np.ndarray): Output values.
            min (np.ndarray): Array of dictionaries, one for each output,
                              containing 'x', 'f' for the minimum of that output.
            max (np.ndarray): Array of dictionaries, one for each output,
                              containing 'x', 'f' for the maximum of that output.
            bounds (np.ndarray): 2D array of lower and upper bounds for each output.

    notes:
        - use `foo.un` to access the UncertainNumber objects.
    """

    def __init__(self, un: np.ndarray = None, raw_data: dict = None):
        if un is None:
            # Initialize as an empty array if un is None
            self.un = np.array([])
        else:
            self.un = un

        if raw_data is not None:
            self.raw_data = raw_data
        else:
            self.raw_data = {
                "x": None,
                "f": None,
                "min": np.array([]),
                "max": np.array([]),
                "bounds": np.array([]),
            }

    def add_raw_data(self, x=None, f=None, K=None, sign_x: np.ndarray = None):
        """Adds raw data to the results."""
        if x is not None:
            self.raw_data["x"] = x
        if K is not None:
            self.raw_data["K"] = K
        if f is not None:
            self.raw_data["f"] = f
        if sign_x is not None:
            if (
                isinstance(sign_x, np.ndarray) and len(sign_x.shape) > 1
            ):  # Multiple outputs
                self.raw_data["sign_x"] = sign_x
            else:  # Single output
                self.raw_data["sign_x"] = np.array([sign_x])  # Wrap in an array

    def summary(self):
        """Prints the results in a formatted way, handling None values and multiple outputs."""

        if self.raw_data["f"] is not None:  # Check if 'f' exists and is not None
            if len(self.raw_data["f"].shape) == 1:  # 1D array, single output
                num_outputs = 1
            else:  # 2D array, multiple outputs
                num_outputs = self.raw_data["f"].shape[1]  # Number of columns
        # Check if 'bounds' exists and is not None (Corrected)
        elif self.raw_data["bounds"] is not None and len(self.raw_data["bounds"]) > 0:
            if len(self.raw_data["bounds"].shape) == 3:  # 1D array, single output
                num_outputs = 1
            else:  # 2D array, multiple outputs
                # Number of rows (Corrected)
                num_outputs = self.raw_data["bounds"].shape[0]
        else:
            num_outputs = 1  # Or handle the case where 'f' is None appropriately

        for i in range(num_outputs):
            print("-" * 30)
            print(f"Output {i+1}:")
            print("-" * 30)

            if self.un is not None:
                if num_outputs == 1:
                    print("Uncertain Number:", self.un)
                else:
                    print("Uncertain Number:", self.un[i])
            else:
                print("Uncertain Number: None")

            if (
                "bounds" in self.raw_data
                and self.raw_data["bounds"] is not None
                and len(self.raw_data["bounds"]) > 0
            ):
                print("-" * 30)
                if num_outputs == 1:
                    print("Bounds:", self.raw_data["bounds"])
                else:
                    print("Bounds:", self.raw_data["bounds"][i])

            if (
                "min" in self.raw_data
                and self.raw_data["min"] is not None
                and len(self.raw_data["min"]) > 0
            ):
                print("-" * 30)
                print("Minimum:")
                min_data = self.raw_data["min"][i]
                if min_data.get("f") is not None:
                    print("f:", min_data.get("f"))
                    if min_data.get("x") is None:  # Handle the case where 'x' is None
                        print("x: None")
                    else:
                        print("x:", min_data.get("x"))
                    # Print additional results only for local_optimisation
                    if "final_simplex" in min_data:
                        print("niterations:", min_data.get("niterations"))
                        print("nfevaluations:", min_data.get("nfevaluations"))
                        print("final_simplex:", min_data.get("final_simplex"))
                        print("message:", min_data.get("message"))
                    if "ngenerations" in min_data:
                        print("niterations:", min_data.get("niterations"))
                        print("ngenerations", min_data.get("ngenerations"))
                        print("message:", min_data.get("message"))

            if (
                "max" in self.raw_data
                and self.raw_data["max"] is not None
                and len(self.raw_data["max"]) > 0
            ):
                print("-" * 30)
                print("Maximum:")
                max_data = self.raw_data["max"][i]
                if max_data.get("f") is not None:
                    print("f:", max_data.get("f"))
                    if max_data.get("x") is None:  # Handle the case where 'x' is None
                        print("x: None")
                    else:
                        print("x:", max_data.get("x"))
                    if "final_simplex" in max_data:
                        print("niterations:", max_data.get("niterations"))
                        print("nfevaluations:", max_data.get("nfevaluations"))
                        print("final_simplex:", max_data.get("final_simplex"))
                        print("message:", max_data.get("message"))
                    if "ngenerations" in max_data:
                        print("niterations:", max_data.get("niterations"))
                        print("ngenerations:", max_data.get("ngenerations"))
                        print("message:", max_data.get("message"))

            if "sign_x" in self.raw_data:
                print("-" * 30)
                print("sign_x:", self.raw_data["sign_x"][i])

        print("-" * 30)
        print("Input combinations and corresponding output(s):")
        # Check if 'x' is not None
        if self.raw_data["x"] is not None and len(self.raw_data["x"]) > 0:
            print("x:", self.raw_data["x"])
        else:
            print("x: None")

        if "K" in self.raw_data:  # Check if the keys exist
            print("-" * 30)
            print("K:", self.raw_data["K"])
        print("-" * 30)

        # Check if 'x' is not None
        if self.raw_data["f"] is not None and len(self.raw_data["f"]) > 0:
            print("f:", self.raw_data["f"])  # Print directly if single output
        else:
            print("f: None")
        print("-" * 30)


def process_alea_results(results):
    """
    args:
        - results (Propagation_results): A `Propagation_results` object containing raw
                                epistemic propagation results. This object is
                                modified in-place.

    signature:
        - process_alea_results(results: Propagation_results) -> Propagation_results

    notes:
        - Processes the results of aleatory uncertainty propagation.

        - This function takes a `Propagation_results` object containing raw aleatory
            propagation results and performs the following actions:

            1. Creates `Distribution` objects:
                - If output data exists in `results.raw_data['f']`, it creates an 'UncertainNumber'
                    object  for each output dimension using the sample data.
                - These `UncertainNumber` objects are stored in `results.un`.
                - They have essense = 'distribution'

            2. Saves raw data (optional):
                - If `save_raw_data` is set to 'yes', it saves the raw propagation data
                    (input samples and corresponding output values) to a file.

    returns:
        - Propagation_results: The modified `Propagation_results` object with
                        `UncertainNumber` objects added to `results.un` and
                        potentially with raw data saved to a file.

    raises:
        - ValueError: If the shape of `results.raw_data['f']` is invalid

    examples:
        >>> a = mixed_propagation(vars= [y, L, I, F, E],
        >>>                 fun= cantilever_beam_func,
        >>>                 method= 'monte_carlo',
        >>>                 n_disc=8,
        >>>                 save_raw_data= "no"
        >>>             )
    """
    if results.raw_data["f"] is None:  # Access raw_data from results object
        # UncertainNumber(essence="distribution", distribution_parameters=None, **kwargs)
        results.un = None
    else:
        results.un = []
        # Access raw_data from results object
        for sample_data in results.raw_data["f"].T:
            # results.un.append(UncertainNumber(essence="distribution", distribution_parameters=sample_data, **kwargs))
            results.un.append(Distribution(sample_data=sample_data))

    if save_raw_data == "yes":
        res_path = create_folder(base_path, method)
        save_results(results.raw_data, method=method, res_path=res_path, fun=fun)

    return results


def process_results(results: Propagation_results):
    """
    args:
        - results (Propagation_results): A `Propagation_results` object containing raw
                                epistemic propagation results. This object is
                                modified in-place.

    notes:
        - Processes the results of epistemic uncertainty propagation.

        - This function takes a `Propagation_results` object containing raw epistemic
            propagation results and performs the following actions:

            1. Creates `UncertainNumber` objects:
                - If output bounds exist in `results.raw_data['bounds']`, it creates
                    `UncertainNumber` objects with "interval" essence, representing the
                    resulting interval uncertainty.
                - It handles both single-output (1D array of bounds) and multi-output
                    (2D array of bounds) cases.
                - These `UncertainNumber` objects are stored in `results.un`.

            2. Saves raw data (optional):
                - If `save_raw_data` is set to 'yes', it saves the raw propagation data
                    to a file.

    signature:
        - process_results(results: Propagation_results) -> Propagation_results

    returns:
        - Propagation_results: The modified `Propagation_results` object with
                        `UncertainNumber` objects added to `results.un` and
                        potentially with raw data saved to a file.

    raises:
        - ValueError: If the shape of `results.raw_data['bounds']` is invalid.

    """
    if results.raw_data["bounds"] is None or results.raw_data["bounds"].size == 0:
        results.un = UncertainNumber(essence="interval", bounds=None, **kwargs)
    else:
        if results.raw_data["bounds"].ndim == 2:  # 2D array
            results.un = [
                UncertainNumber(essence="interval", bounds=bound, **kwargs)
                for bound in results.raw_data["bounds"]
            ]
        # 1D array
        elif (
            results.raw_data["bounds"].ndim == 1
            and len(results.raw_data["bounds"]) == 2
        ):
            results.un = UncertainNumber(
                essence="interval", bounds=results.raw_data["bounds"], **kwargs
            )
        else:
            raise ValueError(
                "Invalid shape for 'bounds'. Expected 2D array or 1D array with two values."
            )

    if save_raw_data == "yes":
        res_path = create_folder(base_path, method)
        save_results(results.raw_data, method=method, res_path=res_path, fun=fun)

    return results

    def process_mixed_results(results: Propagation_results):
        """
        args:
            - results (Propagation_results): A `Propagation_results` object containing raw
                                    epistemic propagation results. This object is
                                    modified in-place.

        signature:
            - process_mixed_results(results: Propagation_results) -> Propagation_results

        notes:
            - Processes the results of mixed uncertainty propagation.

            - This function takes a `Propagation_results` object containing raw aleatory
              propagation results and performs the following actions:

                1. Creates `UncertainNumber` objects:
                    - If output data exists in `results.raw_data['bounds']`, it creates an 'UncertainNumber'
                      object  for each output dimension using the sample data.
                    - These `UncertainNumber` objects are stored in `results.un`.
                    - The `UncertainNumber` has essense = 'pbox'.

                2. Saves raw data (optional):
                    - If `save_raw_data` is set to 'yes', it saves the raw propagation data
                      (input samples and corresponding output values) to a file.

        returns:
            - Propagation_results: The modified `Propagation_results` object with
                          `UncertainNumber` objects added to `results.un` and
                          potentially with raw data saved to a file.

        """
        # if results.raw_data['bounds'] is None or results.raw_data['bounds'].size == 0:
        #     results.un = None
        # else:
        #     if results.raw_data['bounds'].ndim == 4:  # 2D array
        #         results.un = UncertainNumber( essence='interval', bounds=[1, 2])  #[UncertainNumber(essence="pbox", pbox_parameters = bound, **kwargs) for bound in results.raw_data['bounds']]
        #     elif results.raw_data['bounds'].ndim == 3:  # 1D array
        #         results.un =  UncertainNumber( essence='interval', bounds=[1, 2]) #UncertainNumber(essence="pbox",  pbox_parameters=results.raw_data['bounds'], **kwargs)
        #     else:
        #         raise ValueError("Invalid shape for 'bounds'. Expected 2D array or 1D array with two values.")

        # if save_raw_data == "yes":
        # res_path = create_folder(base_path, method)
        # save_results(results.raw_data, method=method, res_path=res_path, fun=fun)

        return results


# * ------------- more functions ------------- *#


def header_results(all_output, all_input, method=None):
    """
    Determine generic header for output and input.

    Args:
        all_output (np.ndarray): A NumPy array containing the output values.
        all_input (np.ndarray): A NumPy array containing the input values.

    Returns:
        list: A list of strings representing the header for the combined DataFrame.
    """
    if all_output is None:
        header_y = []
    else:
        if all_output.ndim == 1:
            len_y = 1
        else:
            len_y = all_output.shape[1]
        header_y = ["y" + str(i) for i in range(len_y)]

    if method in "cauchy":
        m = all_input.shape[1] - 1  # Exclude 'K' from the count
        header_x = ["x" + str(i) for i in range(m)] + ["K"]  # Add 'K' at the end
    else:
        m = all_input.shape[1]
        header_x = ["x" + str(i) for i in range(m)]

    header = header_y + header_x
    return header


def post_processing(
    all_input: np.ndarray, all_output: np.ndarray = None, method=None, res_path=None
):
    """Post-processes the results of an uncertainty propagation (UP) method.

    This function takes the input and output values from a UP method, combines them into a
    pandas DataFrame, and optionally saves the raw data to a CSV file. It also checks for
    NaN values in the output and logs them with their corresponding input values if found.
    If all_output is None, it creates a DataFrame with only the input data.

    Args:
        all_input (np.ndarray): A NumPy array containing the input values used in the UP method.
        all_output (np.ndarray, optional): A NumPy array containing the corresponding output
                                            values from the UP method. Defaults to None.
        res_path (str, optional): The path to the directory where the results will be saved.
                                    Defaults to None.

    Returns:
        pandas.DataFrame: A pandas DataFrame containing the combined output and input data
                        (if all_output is provided). If all_output is None, it returns a
                        DataFrame with only the input data.
    """

    if all_output is None:
        print("No function was evaluated. Only input is available")

        df_input = pd.DataFrame(all_input)

        # Handle Cauchy input with 'x' and 'K' (moved outside the if block)
        # if method in ( "endpoints_cauchy"):
        #     x_fields = pd.DataFrame(all_input)  # Get all 'x' field names
        #     print("x_fields", x_fields)
        #     x_data = all_input[x_fields]  # Select only the 'x' fields
        #     df_input = pd.DataFrame(x_data)
        # else:
        #     df_input = pd.DataFrame(all_input)

        header = header_results(all_output=None, all_input=all_input, method=method)
        df_input.columns = header
        df_output_input = df_input

    else:
        # Transform np.array input-output into pandas data.frame
        df_input = pd.DataFrame(all_input)
        df_output = pd.DataFrame(all_output)

        # Create a single output input data.frame
        df_output_input = pd.concat([df_output, df_input], axis=1)

        # determine generic header for output and input
        header = header_results(all_output, all_input, method)
        df_output_input.columns = header

    # Return .csv with raw data only if asked ###
    if res_path is not None:
        create_csv(res_path, "Raw_data.csv", df_output_input)

    # Check for NaN values ONLY if all_output is provided
    if all_output is not None:
        df_NA = df_output_input[df_output_input.isna().any(axis=1)]
        if len(df_NA) != 0:
            # The input values are rounded to ensure equality
            df_NA = df_NA.apply(np.round, args=[4])
            df_NA_unique = df_NA.drop_duplicates(keep="first", ignore_index=True)

            create_csv(res_path, "NAlog.csv", df_NA_unique)
        else:
            print("There are no NA values produced by the input")

    return df_output_input


def create_folder(base_path, method):
    """Creates a folder named after the called UP method where the results files are stored

    args:
        - base_path: The base path
        - method: the name of the called method

    signature:
        create_folder(base_path: string, method: string ) -> path.folder

    note:
        - the augument `base_path` will specify the location of the created results folder.
        - the argument `method` will provide the name for the results folder.

    return:
        -  A folder in a prespecified path

    example:
        base_path = "C:/Users/DAWS2_code/UP"
        method = "vertex"
        y = create_folder(base_path, method)
    """
    base_path = Path(base_path)

    res_path = base_path / method
    res_path.mkdir(parents=True, exist_ok=True)

    return res_path


def create_csv(res_path, filename, data):
    """Creates a .csv file and sotres it in a pre-specified folder with results generated by a UP method

    args:
        - res_path: A folder in a prespecified path named after the called UP method
        - filename: the name of the file
        - data: a pandas.dataframe with results from UP method

    signature:
        create_csv(res_path = path, filename = filename, data = pandas.dataframe) -> path.filename

    note:
        - the augument `res_path` will specify the folder where the .csv file will be created.
        - argument `file` will provide the name of hte .csv file.
        - argument `data` will provide data in terms of pandas.dataframe.

    return:
        -  A .csv file in a prespecified folder

    example:
        base_path = "C:/Users/DAWS2_code/UP/vertex"
        filename = 'min_max_values'
       df = pd.DataFrame(
         {"Name" : ["y0", "y0"],
          "fun"  : ["min","max"]
          "y0" : [4, 6]}, index = [1, 2, 3])
        header = ['Name', 'fun', 'values']
        y = create_csv(res_path, filename, df)
    """
    try:
        # Attempt to open the file
        file_path = res_path / filename
        with open(file_path, "w", newline="") as f:
            writer = csv.writer(f)
            writer.writerow(data.columns)
            writer.writerows(data.values.tolist())
    except FileNotFoundError:
        print("The file does not exist.")

    return filename


def save_results(data, method, res_path, fun=None):
    if fun is None:
        Results = post_processing(
            data["x"], all_output=None, method=method, res_path=res_path
        )
    else:
        Results = post_processing(data["x"], data.get("f"), method, res_path)

    return Results


def condense_bounds(bounds, N):
    """
    Condenses lower and upper bounds of a probability distribution to a specified size.

    Args:
      bounds: A NumPy array of shape (num_outputs, 2, num_points) representing the lower
              and upper bounds of a probability distribution for potentially multiple outputs.
              The first dimension corresponds to different outputs of the function,
              the second dimension corresponds to lower and upper bounds (0 for lower, 1 for upper),
              and the third dimension corresponds to the original discretization points.
      N: The desired size of the condensed arrays.

    Returns:
      A NumPy array of shape (num_outputs, 2, N) containing the condensed lower and upper bounds.
    """
    num_outputs = bounds.shape[0]
    num_points = bounds.shape[2]

    # Handle different condensation sizes for each output
    if isinstance(N, int):
        N = [N] * num_outputs  # Create a list with the same size for all outputs

    # Initialize with the maximum size
    condensed_bounds = np.zeros((num_outputs, 2, max(N)))

    for i in range(num_outputs):
        interval_size = num_points // N[i]

        lower_bounds_sorted = np.sort(bounds[i, 0, :])
        upper_bounds_sorted = np.sort(bounds[i, 1, :])

        condensed_lower = np.array(
            [
                lower_bounds_sorted[j * interval_size + interval_size - 1]
                for j in range(N[i])
            ]
        )
        condensed_upper = np.array(
            [upper_bounds_sorted[j * interval_size] for j in range(N[i])]
        )

        # Assign to the correct slice
        condensed_bounds[i, 0, : N[i]] = condensed_lower
        condensed_bounds[i, 1, : N[i]] = condensed_upper

    return condensed_bounds



import numpy as np
from typing import Callable, Union
import matplotlib.pyplot as plt

from .epistemic_uncertainty.endpoints import endpoints_method
from .epistemic_uncertainty.extremepoints import extremepoints_method
from .epistemic_uncertainty.subinterval import subinterval_method
from .epistemic_uncertainty.sampling import sampling_method
from .epistemic_uncertainty.genetic_optimisation import genetic_optimisation_method
from .local_optimisation import local_optimisation_method
from .epistemic_uncertainty.endpoints_cauchy import cauchydeviates_method
from .aleatory_uncertainty.sampling_aleatory import sampling_aleatory_method
from .mixed_uncertainty.second_order_propagation import second_order_propagation_method
from .mixed_uncertainty.first_order_propagation import first_order_propagation_method
from .utils import (
    create_folder,
    save_results,
    Propagation_results,
)
from pyuncertainnumber.characterisation.uncertainNumber import (
    UncertainNumber,
    Distribution,
)


# TODO the cauchy with save_raw_data = 'yes' raises issues.
# TODO update the descriptions and code for process_results once more.
""" the old top-level UP implementation """

# * ---------------------the top level UP function ---------------------*#


def aleatory_propagation(
    vars: list = None,
    results: Propagation_results = None,
    func: Callable = None,
    n_sam: int = 500,
    method: str = "monte_carlo",
    save_raw_data=False,
    *,  # Keyword-only arguments start here
    base_path=np.nan,
    **kwargs,
):
    """This function propagates aleatory uncertainty through a given function (`func`) using either Monte Carlo or Latin Hypercube sampling, considering the aleatory uncertainty represented by a list of `UncertainNumber` objects (`vars`).
    args:
        - vars (list): A list of UncertainNumber objects, each representing an input
                    variable with its associated uncertainty.
        - func (Callable): The function to propagate uncertainty through.
        - n_sam (int): The number of samples to generate.
                    Default is 500.
        - method (str, optional): The sampling method ('monte_carlo' or 'latin_hypercube').
                    Defaults to 'monte_carlo'.
        - save_raw_data (str, optional): Whether to save raw data ('yes' or 'no').
                    Defaults to 'no'.
        - base_path (str, optional): Path for saving results (if save_raw_data is 'yes').
                    Defaults to np.nan.
        - **kwargs: Additional keyword arguments to be passed to the UncertainNumber constructor.

    signature:
        aleatory_propagation(x:np.ndarray, f:Callable, n:int, ...) -> Propagation_results

    note:
        - If the `f` function returns multiple outputs, the `all_output` array will be 2-dimensional y and x for all x samples.

    returns:
        Propagation_results: A  `Propagation_results` object containing:
                          - 'un': A list of UncertainNumber objects, each representing
                                  the output(s) of the function.
                          - 'raw_data': A dictionary containing raw data (if
                                        save_raw_data is 'yes'):
                                          - 'x': All generated input samples.
                                          - 'f': Corresponding output values for each
                                                input sample.

    raises:
        ValueError: For invalid method, save_raw_data, or missing arguments.

    """
    # Input validation

    match method:
        case "monte_carlo" | "latin_hypercube":
            assert n_sam is not None, "number of samples) is required for sampling "

            results = sampling_aleatory_method(
                vars,
                func,
                results,
                n_sam,
                method=method.lower(),
                save_raw_data=save_raw_data,
            )
            return process_alea_results(results)
        case "taylor_expansion":
            print("Taylor expansion is not implemented in this version")
        case _:
            raise ValueError("Invalid UP method.")


def mixed_propagation(
    vars: list,
    func: Callable = None,
    results: Propagation_results = None,
    method="second_order_extremepoints",
    n_disc: Union[int, np.ndarray] = 10,
    condensation: int = None,
    tOp: Union[float, np.ndarray] = 0.999,
    bOt: Union[float, np.ndarray] = 0.001,
    save_raw_data="no",
    *,  # Keyword-nly arguments start here
    base_path=np.nan,
    **kwargs,
):
    """Performs mixed uncertainty propagation through a given function. This function handles uncertainty propagation when there's a mix of
          aleatory and epistemic uncertainty in the input variables.

    args:
        - vars (list): A list of uncertain variables, which can be a mix of different
                 uncertainty types (e.g., intervals, distributions).
        - func (Callable): The function to propagate uncertainty through.
        - results (Propagation_results, optional): An object to store propagation results.
                                Defaults to None, in which case a new
                                `Propagation_results` object is created.
        - method (str, optional): The mixed uncertainty propagation method. Can be one of:
                                'second_order_endpoints', 'second_order_vertex', 'second_order_extremepoints',
                                'first_order_extremepoints'.
                                Defaults to 'second_order_extremepoints'.
        - n_disc (Union[int, np.ndarray], optional): Number of discretization points for
                                interval variables.
                                Defaults to 10.
        - condensation (int, optional): Parameter for reducing the complexity of the output
                                uncertainty representation.
                                Defaults to None.
        - tOp (Union[float, np.ndarray], optional): Upper threshold or bound used in some methods.
                                Defaults to 0.999.
        - bOt (Union[float, np.ndarray], optional): Lower threshold or bound used in some methods.
                                Defaults to 0.001.
        - save_raw_data (str, optional): Whether to save raw data ('yes' or 'no').
                                Defaults to 'no'.
        - base_path (str, optional): Path for saving results (if save_raw_data is 'yes').
                               Defaults to np.nan.
        - **kwargs: Additional keyword arguments passed to the underlying propagation methods.

    signature:
       mixed_propagation(vars: list, func: Callable, results: Propagation_results = None, ...) -> Propagation_results

    notes:
        - It can be used if each uncertain number is exrpessed in terms of precise distributions.

    returns:
        Propagation_results: A `Propagation_results` object containing the results of
                          the mixed uncertainty propagation. The format of the results
                          depends on the chosen `method`.

    raises:
        ValueError: For invalid `method` or `save_raw_data`.

    examples:
        >>> a = mixed_propagation(vars= [y, L, I, F, E],
                            func= cantilever_beam_func,
                            method= 'second_order_extremepoints',
                            n_disc=8,
                            #save_raw_data= "no"#,
                            save_raw_data= "yes",
                            base_path= base_path
                        )
    """

    if save_raw_data not in ("yes", "no"):  # Input validation
        raise ValueError("Invalid save_raw_data option. Choose 'yes' or 'no'.")

    match method:
        case "second_order_endpoints" | "second_order_vertex" | "endpoints" | "vertex":
            results = second_order_propagation_method(
                vars,
                func,
                results,
                method="endpoints",
                n_disc=n_disc,
                condensation=condensation,
                tOp=tOp,
                bOt=bOt,
                save_raw_data=save_raw_data,
                **kwargs,
            )  # Pass save_raw_data directly
            return process_mixed_results(results)

        case "second_order_extremepoints" | "extremepoints":
            results = second_order_propagation_method(
                vars,
                func,
                results,
                method="extremepoints",
                n_disc=n_disc,
                condensation=condensation,
                tOp=tOp,
                bOt=bOt,
                save_raw_data=save_raw_data,
                **kwargs,
            )  # Pass save_raw_data directly
            return process_mixed_results(results)

        case "first_order" | "first_order_extremepoints":
            results = first_order_propagation_method(
                vars,
                func,
                results,
                # method = 'extremepoints',
                n_disc=n_disc,
                condensation=condensation,
                tOp=tOp,
                bOt=bOt,
                save_raw_data=save_raw_data,
                **kwargs,
            )
            return process_mixed_results(results)
        case _:
            raise ValueError("Invalid UP method.")


def epistemic_propagation(
    vars,
    func,
    method: str = None,
    save_raw_data=False,
    *,
    results: Propagation_results = None,
    n_sub: np.integer = None,
    n_sam: np.integer = None,
    x0: np.ndarray = None,
    base_path=np.nan,
    tol_loc: np.ndarray = None,
    options_loc: dict = None,
    method_loc="Nelder-Mead",
    pop_size=1000,
    n_gen=100,
    tol=1e-3,
    n_gen_last=10,
    algorithm_type="NSGA2",
    **kwargs,
):
    """Performs epistemic uncertainty propagation through a given function. This function implements various methods for propagating epistemic uncertainty,
          typically represented as intervals.

    args:
        - vars (list): A list of `UncertainNumber` objects representing the input variables
                 with their associated interval uncertainty.
        - func (Callable): The function to propagate uncertainty through.
        - results (Propagation_results, optional): An object to store propagation results.
                        Defaults to None, in which case a new
                        `Propagation_results` object is created.
        - n_sub (np.integer, optional): Number of subintervals for subinterval methods.
                        Defaults to None.
        - n_sam (np.integer, optional): Number of samples for sampling-based methods.
                        Defaults to None.
        - x0 (np.ndarray, optional): Initial guess for local optimization methods.
                        Defaults to None.
        - method (str, optional): The uncertainty propagation method to use.
                        Defaults to "endpoint".
        - save_raw_data (str, optional): Whether to save raw data ('yes' or 'no').
                        Defaults to "no".
        - base_path (str, optional): Path for saving results (if save_raw_data is 'yes').
                        Defaults to np.nan.
        - tol_loc (np.ndarray, optional): Tolerance for local optimization.
                        Defaults to None.
        - options_loc (dict, optional): Options for local optimization.
                        Defaults to None.
        - method_loc (str, optional): Method for local optimization.
                        Defaults to 'Nelder-Mead'.
        - pop_size (int, optional): Population size for genetic algorithms.
                        Defaults to 1000.
        - n_gen (int, optional): Number of generations for genetic algorithms.
                        Defaults to 100.
        - tol (float, optional): Tolerance for genetic algorithms. Defaults to 1e-3.
        - n_gen_last (int, optional): Number of last generations for genetic algorithms.
                        Defaults to 10.
        - algorithm_type (str, optional): Type of genetic algorithm.
                        Defaults to 'NSGA2'.
        - **kwargs: Additional keyword arguments passed to the `UncertainNumber` constructor.

    signature:
        epistemic_propagation(vars: list, func: Callable, results: Propagation_results = None, ...) -> Propagation_results

    notes:
        -  It supports a wide range of techniques, including:

            1. Interval-based methods:
                - `endpoints` or `vertex`:  Calculates the function output at the endpoints
                                 or vertices of the input intervals.
                - `extremepoints`: Considers all possible combinations of interval endpoints
                       to find the extreme values of the output.
                - `subinterval` or `subinterval_reconstitution`: Divides the input intervals
                                                    into subintervals and performs
                                                    propagation on each subinterval.

            2. Sampling-based methods:
                - `monte_carlo`, `latin_hypercube`:  Uses Monte Carlo or Latin Hypercube
                                        sampling within the input intervals.
                - `monte_carlo_endpoints`, `latin_hypercube_endpoints`:  Combines sampling with
                                                            evaluation at interval
                                                            endpoints.
                - `cauchy`, `endpoint_cauchy`, `endpoints_cauchy`: Uses Cauchy deviates for
                                                      sampling.

            3. Optimization-based methods:
                - `local_optimization` or `local_optimisation`:  Uses local optimization
                                                      algorithms to find the minimum
                                                      or maximum output values.
                - `genetic_optimisation` or `genetic_optimization`: Uses genetic algorithms for
                                                        global optimization.
    returns:
        - Propagation_results: A  `Propagation_results` object containing the results of
                          the epistemic uncertainty propagation. The format of the
                          results depends on the chosen `method`.

    raises:
        - ValueError: For invalid `method`, `save_raw_data`, or missing arguments.
        - TypeError: If `func` is not callable for optimization methods.

    example:
        >>> a = epistemic_propagation(vars= [ y, L, I, F, E],
                                func= cantilever_beam_func,
                                method= 'extremepoints',
                                n_disc=8,
                                save_raw_data= "no"
                            )
    """
    # vars = _parse_interverl_inputs(vars)
    x = np.zeros((len(vars), 2))
    for i, un in enumerate(vars):
        x[i, :] = un.bounds  # Get an np.array of bounds for all vars

    match method:
        case "endpoint" | "endpoints" | "vertex":
            # Pass save_raw_data directly
            results = endpoints_method(x, func, results, save_raw_data)

        case "extremepoints":
            # Pass save_raw_data directly
            results = extremepoints_method(x, func, results, save_raw_data)

        case "subinterval" | "subinterval_reconstitution":
            if n_sub is None:
                raise ValueError(
                    "n (number of subintervals) is required for subinterval methods."
                )
            # Pass save_raw_data directly
            results = subinterval_method(x, func, results, n_sub, save_raw_data)

        # case "monte_carlo" | "latin_hypercube":
        #     if n_sam is None:
        #         raise ValueError(
        #             "n (number of samples) is required for sampling methods."
        #         )
        #     results = sampling_method(
        #         x,
        #         func,
        #         results,
        #         n_sam,
        #         method=method.lower(),
        #         endpoints=False,
        #         save_raw_data=save_raw_data,
        #     )

        # case "monte_carlo_endpoints":
        #     if n_sam is None:
        #         raise ValueError(
        #             "n (number of samples) is required for sampling methods."
        #         )
        #     results = sampling_method(
        #         x,
        #         func,
        #         results,
        #         n_sam,
        #         method="monte_carlo",
        #         endpoints=True,
        #         save_raw_data=save_raw_data,
        #     )

        # case "latin_hypercube_endpoints":
        #     if n_sam is None:
        #         raise ValueError(
        #             "n (number of samples) is required for sampling methods."
        #         )
        #     results = sampling_method(
        #         x,
        #         func,
        #         results,
        #         n_sam,
        #         method="latin_hypercube",
        #         endpoints=True,
        #         save_raw_data=save_raw_data,
        #     )

        case "cauchy" | "endpoint_cauchy" | "endpoints_cauchy":
            if n_sam is None:
                raise ValueError(
                    "n (number of samples) is required for sampling methods."
                )
            results = cauchydeviates_method(x, func, results, n_sam, save_raw_data)

        case (
            "local_optimization"
            | "local_optimisation"
            | "local optimisation"
            | "local optimization"
        ):

            if save_raw_data == "yes":
                print("The intermediate steps cannot be saved for local optimisation")
            results = local_optimisation_method(
                x,
                func,
                results,
                x0,
                tol_loc=tol_loc,
                options_loc=options_loc,
                method_loc=method_loc,
            )

        case (
            "genetic_optimisation"
            | "genetic_optimization"
            | "genetic optimization"
            | "genetic optimisation"
        ):
            if save_raw_data == "yes":
                print("The intermediate steps cannot be saved for genetic optimisation")
            results = genetic_optimisation_method(
                x, func, results, pop_size, n_gen, tol, n_gen_last, algorithm_type
            )

        case _:
            raise ValueError("Invalid UP method.")

    return process_results(results)


def Propagation(
    vars: list,
    func: Callable,
    results: Propagation_results = None,
    n_sub: np.integer = 3,
    n_sam: np.integer = 500,
    x0: np.ndarray = None,
    method=None,
    n_disc: Union[int, np.ndarray] = 10,
    condensation: int = None,
    tOp: Union[float, np.ndarray] = 0.999,
    bOt: Union[float, np.ndarray] = 0.001,
    save_raw_data="no",
    *,  # Keyword-only arguments start here
    base_path=np.nan,
    tol_loc: np.ndarray = None,
    options_loc: dict = None,
    method_loc="Nelder-Mead",
    pop_size=1000,
    n_gen=100,
    tol=1e-3,
    n_gen_last=10,
    algorithm_type="NSGA2",
    **kwargs,
):
    """Performs uncertainty propagation through a given function with uncertain inputs. This function automatically selects and executes an appropriate uncertainty propagation method based on the types of uncertainty in the input variables. It supports interval analysis, probabilistic methods, and mixed uncertainty propagation.

    args:
        - vars (list): A list of uncertain variables.
        - func (Callable): The function through which to propagate uncertainty.
        - results (Propagation_results, optional): An object to store propagation results.
                                Defaults to None, in which case a new
                                `Propagation_results` object is created.
        - n_sub (np.integer, optional): Number of subintervals for interval-based methods.
                            Defaults to 3.
        - n_sam (np.integer, optional): Number of samples for Monte Carlo simulation.
                            Defaults to 500.
        - x0 (np.ndarray, optional): Initial guess for optimization-based methods.
                            Defaults to None.
        - method (str, optional):  Specifies the uncertainty propagation method.
                            Defaults to None, which triggers automatic selection.
        - n_disc (Union[int, np.ndarray], optional): Number of discretization points.
                            Defaults to 10.
        - condensation (int, optional): Parameter for reducing output complexity.
                            Defaults to None.
        - tOp (Union[float, np.ndarray], optional): Upper threshold or bound.
                            Defaults to 0.999.
        - bOt (Union[float, np.ndarray], optional): Lower threshold or bound.
                            Defaults to 0.001.
        - save_raw_data (str, optional): Whether to save intermediate results ('yes' or 'no').
                            Defaults to 'no'.
        - base_path (str, optional): Path for saving data. Defaults to np.nan.
        - tol_loc (np.ndarray, optional): Tolerance for local optimization.
                            Defaults to None.
        - options_loc (dict, optional): Options for local optimization.
                            Defaults to None.
        - method_loc (str, optional): Method for local optimization.
                            Defaults to 'Nelder-Mead'.
        - pop_size (int, optional): Population size for genetic algorithms.
                            Defaults to 1000.
        - n_gen (int, optional): Number of generations for genetic algorithms.
                           Defaults to 100.
        - tol (float, optional): Tolerance for genetic algorithms. Defaults to 1e-3.
        - n_gen_last (int, optional): Number of last generations for genetic algorithms.
                            Defaults to 10.
        - algorithm_type (str, optional): Type of genetic algorithm.
                            Defaults to 'NSGA2'.
        **kwargs: Additional keyword arguments passed to the underlying propagation methods.

    signature:
       - Propagation(vars: list, func: Callable, results: Propagation_results = None, ...) -> Propagation_results


    return:
        - Propagation_results: A  `Propagation_results` object including:
                        - 'un': A list of UncertainNumber objects, each representing
                                  the output(s) of the function.
                        - 'raw_data': depending on the method selected.

    example:

        >>> a = Propagation(vars= [ y, L, I, F, E],
                            func= cantilever_beam_func,
                            method= 'extremepoints',
                            n_disc=8,
                            save_raw_data= "no"
                        )
    """
    essences = [un.essence for un in vars]  # Get a list of all essences

    if results is None:
        results = Propagation_results()  # Create an instance of Propagation_results

    # Determine the plotting strategy based on essences
    if all(essence == "interval" for essence in essences):

        y = epistemic_propagation(
            vars=vars,
            func=func,
            results=results,
            n_sub=n_sub,
            n_sam=n_sam,
            x0=x0,
            method=method,
            save_raw_data=save_raw_data,
            base_path=base_path,
            tol_loc=tol_loc,
            options_loc=options_loc,
            method_loc=method_loc,
            pop_size=pop_size,
            n_gen=n_gen,
            tol=tol,
            n_gen_last=n_gen_last,
            algorithm_type=algorithm_type,
            **kwargs,
        )

    elif all(essence == "distribution" for essence in essences):
        if method in (
            "second_order_endpoints",
            "second_order_vertex",
            "second_order_extremepoints",
        ):
            y = mixed_propagation(
                vars=vars,
                func=func,
                results=results,
                n_disc=n_disc,
                condensation=condensation,
                tOp=tOp,
                bOt=bOt,
                save_raw_data=save_raw_data,
                base_path=base_path,
                **kwargs,
            )
        else:  # Use aleatory propagation if method is not in the list above
            y = aleatory_propagation(
                vars=vars,
                func=func,
                results=results,
                n_sam=n_sam,
                method=method,
                save_raw_data=save_raw_data,
                base_path=base_path,
                **kwargs,
            )
    else:  # Mixed case or at least one p-box
        y = mixed_propagation(
            vars=vars,
            func=func,
            results=results,
            n_disc=n_disc,
            condensation=condensation,
            tOp=tOp,
            bOt=bOt,
            save_raw_data=save_raw_data,
            base_path=base_path,
            **kwargs,
        )

    return y


def plotPbox(xL, xR, p=None):
    """
    Plots a p-box (probability box) using matplotlib.

    Args:
        xL (np.ndarray): A 1D NumPy array of lower bounds.
        xR (np.ndarray): A 1D NumPy array of upper bounds.
        p (np.ndarray, optional): A 1D NumPy array of probabilities corresponding to the intervals.
                                   Defaults to None, which generates equally spaced probabilities.
        color (str, optional): The color of the plot. Defaults to 'k' (black).
    """
    xL = np.squeeze(xL)  # Ensure xL is a 1D array
    xR = np.squeeze(xR)  # Ensure xR is a 1D array

    if p is None:
        # p should have one more element than xL/xR
        p = np.linspace(0, 1, len(xL))

    if p.min() > 0:
        p = np.concatenate(([0], p))
        xL = np.concatenate(([xL[0]], xL))
        xR = np.concatenate(([xR[0]], xR))

    if p.max() < 1:
        p = np.concatenate((p, [1]))
        xR = np.concatenate((xR, [xR[-1]]))
        xL = np.concatenate((xL, [xL[-1]]))

    colors = "black"
    # Highlight the points (xL, p)
    plt.scatter(xL, p, color=colors, marker="o", edgecolors="black", zorder=3)

    # Highlight the points (xR, p)
    plt.scatter(xR, p, color=colors, marker="o", edgecolors="black", zorder=3)

    plt.fill_betweenx(p, xL, xR, color=colors, alpha=0.5)
    plt.plot([xL[0], xR[0]], [0, 0], color=colors, linewidth=3)
    plt.plot([xL[-1], xR[-1]], [1, 1], color=colors, linewidth=3)
    plt.show()


def main():
    """implementation of any method for epistemic uncertainty on the cantilever beam example"""

    # y = np.array([0.145, 0.155])  # m

    # L = np.array([9.95, 10.05])  # m

    # I = np.array([0.0003861591, 0.0005213425])  # m**4

    # F = np.array([11, 37])  # kN

    # E = np.array([200, 220])  # GPa

    # # Create a 2D np.array with all uncertain input parameters in the **correct** order.
    # xInt = np.array([L, I, F, E])

    def cantilever_beam_deflection(x):
        """Calculates deflection and stress for a cantilever beam.

          Args:
            x (np.array): Array of input parameters:
                x[0]: Length of the beam (m)
                x[1]: Second moment of area (mm^4)
                x[2]: Applied force (N)
                x[3]: Young's modulus (MPa)

        Returns:
            float: deflection (m)
                   Returns np.nan if calculation error occurs.
        """
        beam_length = x[0]
        I = x[1]
        F = x[2]
        E = x[3]
        try:  # try is used to account for cases where the input combinations leads to error in func due to bugs
            deflection = F * beam_length**3 / (3 * E * 10**6 * I)  # deflection in m

        except:
            deflection = np.nan

        return np.array([deflection])

    def cantilever_beam_func(x):

        y = x[0]
        beam_length = x[1]
        I = x[2]
        F = x[3]
        E = x[4]
        try:  # try is used to account for cases where the input combinations leads to error in func due to bugs
            deflection = F * beam_length**3 / (3 * E * 10**6 * I)  # deflection in m
            stress = F * beam_length * y / I / 1000  # stress in MPa

        except:
            deflection = np.nan
            stress = np.nan

        return np.array([deflection, stress])

    #     # example
    # y = UncertainNumber(name='distance to neutral axis', symbol='y', units='m', essence='distribution', distribution_parameters=["gaussian", [0.15, 0.00333]])
    # L = UncertainNumber(name='beam length', symbol='L', units='m', essence='distribution', distribution_parameters=["gaussian", [10.05, 0.033]])
    # I = UncertainNumber(name='moment of inertia', symbol='I', units='m', essence='distribution', distribution_parameters=["gaussian", [0.000454, 4.5061e-5]])
    # F = UncertainNumber(name='vertical force', symbol='F', units='kN', essence='distribution', distribution_parameters=["gaussian", [24, 8.67]])
    # E = UncertainNumber(name='elastic modulus', symbol='E', units='GPa', essence='distribution', distribution_parameters=["gaussian", [210, 6.67]])

    y = UncertainNumber(
        name="beam width",
        symbol="y",
        units="m",
        essence="interval",
        bounds=[0.145, 0.155],
    )
    L = UncertainNumber(
        name="beam length",
        symbol="L",
        units="m",
        essence="interval",
        bounds=[9.95, 10.05],
    )
    I = UncertainNumber(
        name="moment of inertia",
        symbol="I",
        units="m",
        essence="interval",
        bounds=[0.0003861591, 0.0005213425],
    )
    F = UncertainNumber(
        name="vertical force",
        symbol="F",
        units="kN",
        essence="interval",
        bounds=[11, 37],
    )
    E = UncertainNumber(
        name="elastic modulus",
        symbol="E",
        units="GPa",
        essence="interval",
        bounds=[200, 220],
    )

    METHOD = "extremepoints"
    base_path = (
        "C:\\Users\\Ioanna\\OneDrive - The University of Liverpool\\DAWS2_code\\UP\\"
    )

    a = Propagation(
        vars=[L, I, F, E],
        func=cantilever_beam_deflection,
        method="extremepoints",
        n_disc=8,
        # save_raw_data= "no"#,
        save_raw_data="yes",
        base_path=base_path,
    )

    a.print()
    # plotPbox(a.raw_data['min'][0]['f'], a.raw_data['max'][0]['f'], p=None)
    # plt.show()
    # plotPbox(a.raw_data['min'][1]['f'], a.raw_data['max'][1]['f'], p=None)
    # plt.show()

    return a


if __name__ == "__main__":
    main()



import pandas as pd
import numpy as np
from scipy.stats import multivariate_normal, uniform
from scipy.stats import beta, norm, multivariate_normal, truncnorm

# utils_bayesian.py
import numpy as np
from typing import List

class Prior:
    """Abstract prior interface (minimal)."""
    def generate_rns(self, n: int) -> np.ndarray:
        """Return n samples ~ prior (shape: (n, d_i) or (n,))."""
        raise NotImplementedError
    def log_pdf_eval(self, x: np.ndarray) -> float:
        """Return log p(x) for a single vector x."""
        raise NotImplementedError

class PriorUniform(Prior):
    """Axis-aligned uniform prior over [lb, ub] in R^d (vector bounds)."""
    def __init__(self, lb: np.ndarray, ub: np.ndarray, name: str = "uniform"):
        self.lb = np.asarray(lb, float)
        self.ub = np.asarray(ub, float)
        assert self.lb.shape == self.ub.shape
        self.d = self.lb.size
        self.name = name
        self._logZ = -np.sum(np.log(self.ub - self.lb))

    def generate_rns(self, n: int) -> np.ndarray:
        """Vectorized sampling over the hyper-rectangle."""
        return self.lb + (self.ub - self.lb) * np.random.rand(n, self.d)

    def log_pdf_eval(self, x: np.ndarray) -> float:
        """log 1/Vol if x within bounds; -inf otherwise."""
        x = np.asarray(x, float).ravel()
        if np.all(x >= self.lb) and np.all(x <= self.ub):
            return self._logZ
        return -np.inf

def initial_population(n: int, priors: List[Prior]) -> np.ndarray:
    """
    """  # compact: draw n particles across concatenated priors
    blocks = [p.generate_rns(n) for p in priors]
    # Each block may be (n,) or (n,d_i); ensure 2D then hstack
    blocks = [b.reshape(n, -1) for b in blocks]
    return np.hstack(blocks)

def log_prior(x: np.ndarray, priors: List[Prior]) -> float:
    """
    """  # compact: sum component prior logs for a single particle
    x = np.asarray(x, float).ravel()
    logs = []
    off = 0
    for p in priors:
        d = getattr(p, "d", 1)
        logs.append(p.log_pdf_eval(x[off:off+d]))
        off += d
    return float(np.sum(logs))

def log_prior_batch(samples: np.ndarray, priors: List[Prior]) -> np.ndarray:
    """
    """  # compact: batch version of log_prior
    samples = np.atleast_2d(samples)
    n, D = samples.shape
    out = np.zeros(n, float)
    off = 0
    for p in priors:
        d = getattr(p, "d", 1)
        # evaluate each row on its slice
        sl = samples[:, off:off+d]
        out += np.array([p.log_pdf_eval(sl[i]) for i in range(n)], float)
        off += d
    return out

def compute_beta_update_evidence(beta: float,
                                 log_lik: np.ndarray,
                                 log_evidence: float,
                                 prev_ESS: float,
                                 target_frac: float = 0.95,
                                 min_particles: int = 50,
                                 max_iter: int = 30) -> tuple:
    """
    """  # compact: binary-search next beta to hit target ESS ~ target_frac * prev_ESS
    old_beta = beta
    low, high = beta, 1.0
    target = max(target_frac * prev_ESS, min_particles)
    N = log_lik.size
    # numeric stabilize
    ll_max = float(log_lik.max())
    new_beta, Wn, ESS = beta, np.ones(N)/N, prev_ESS
    for _ in range(max_iter):
        mid = 0.5 * (low + high)
        inc = mid - old_beta
        W = np.exp(inc * (log_lik - ll_max))
        Wn = W / (W.sum() + 1e-300)
        ESS = 1.0 / np.sum(Wn**2)
        if abs(ESS - target) < 1.0:
            new_beta = mid
            break
        if ESS < target:
            high = mid
        else:
            low = mid
        new_beta = mid
        if (high - low) < 1e-8:
            break
    if new_beta >= 1.0:  # clamp & recompute at 1.0
        new_beta = 1.0
        inc = new_beta - old_beta
        W = np.exp(inc * (log_lik - ll_max))
        Wn = W / (W.sum() + 1e-300)
        ESS = 1.0 / np.sum(Wn**2)
    # evidence update (log of mean unnormalized weights)
    log_evidence += np.log(W.mean() + 1e-300)
    return new_beta, log_evidence, Wn, ESS

def gaussian_proposals(current: np.ndarray, cov: np.ndarray, n: int) -> np.ndarray:
    """
    """  # compact: draw n proposals ~ N(current, cov)
    D = current.size
    return current + np.random.multivariate_normal(np.zeros(D), cov, size=n)



class JointBivariatePDF:
    """Base class for 2D aleatoric samplers with copulas."""
    def __init__(self, params):
        self.eps = 1e-9
        self.params = params

    def sample(self, ns: int = 1000):
        raise NotImplementedError("implement sampling here.")


class BetaGaussCopulaPDF(JointBivariatePDF):
    """
    Aleatoric sampler: (xa1, xa2) ~ Beta marginals + Gaussian copula.
    Params: [a1, a2, b1, b2, rho] with values in [0, 1].
    Scaled internally via factor `s` (default: 20).
    """
    def __init__(self, params, s=20):
        super().__init__(params)
        a1, a2, b1, b2, rho = params
        self.a1, self.a2 = a1 * s, a2 * s
        self.b1, self.b2 = b1 * s, b2 * s
        self.rho = rho * 2 - 1  # map [0,1] → [-1,1]
        self.cov = np.array([[1.0, self.rho], [self.rho, 1.0]])

    def sample(self, ns: int = 1000):
        z = multivariate_normal(mean=[0, 0], cov=self.cov).rvs(size=ns)
        u = np.clip(norm.cdf(z), self.eps, 1 - self.eps)
        xa1 = beta(self.a1, self.a2).ppf(u[:, 0])
        xa2 = beta(self.b1, self.b2).ppf(u[:, 1])
        return np.stack([xa1, xa2], axis=1)


class NormalGaussCopulaPDF(JointBivariatePDF):
    """ Aleatoric sampler: (xa1, xa2) ~ Truncated Normal marginals """
    def __init__(self, params, s_mu=2, s_sigma=2):
        super().__init__(params)
        mu1, sigma1, mu2, sigma2, rho = params
        # Scale parameters to their actual domain
        self.mu1, self.sigma1 = (mu1 *2 - 1) * s_mu, sigma1 * s_sigma
        self.mu2, self.sigma2 = (mu2 * 2 - 1) * s_mu, sigma2 * s_sigma
        self.rho = rho * 2 - 1  # map [0,1] → [-1,1]
        self.cov = np.array([[1.0, self.rho], [self.rho, 1.0]])

    def sample(self, ns: int = 1000):
        # Step 1: Draw from Gaussian copula
        z = multivariate_normal(mean=[0, 0], cov=self.cov).rvs(size=ns)
        u = np.clip(truncnorm.cdf(z, -5, 5), self.eps, 1 - self.eps)

        # Step 2: Truncated normal marginals on [0,1]
        a1, b1 = (0 - self.mu1) / self.sigma1, (1 - self.mu1) / self.sigma1
        a2, b2 = (0 - self.mu2) / self.sigma2, (1 - self.mu2) / self.sigma2

        x1 = truncnorm.ppf(u[:, 0], a=a1, b=b1, loc=self.mu1, scale=self.sigma1)
        x2 = truncnorm.ppf(u[:, 1], a=a2, b=b2, loc=self.mu2, scale=self.sigma2)

        return np.stack([x1, x2], axis=1)  # shape (ns, 2)


class BetaMixtureGaussCopulaPDF(JointBivariatePDF):
    """
    Mixture of Beta marginals with Gaussian copula correlation.
    Each component uses (a1, a2), (b1, b2) marginals + rho.
    Params: flat vector of length n_param * n_mix with values in [0, 1]; scaled internally.
    """

    def __init__(self, params, n_mix: int =2, s: int =20, weights=None):
        super().__init__(params)
        self.n_mix = n_mix
        self.s = s
        self.eps = 1e-6  # numerical stability

        expected_len = n_mix * 5
        assert len(params) == expected_len, f"Expected {expected_len} parameters for n_mix={n_mix}"

        self.components = []
        for i in range(n_mix):
            a1, a2, b1, b2, rho = params[5 * i: 5 * (i + 1)]
            comp = {
                'a1': a1 * s,
                'a2': a2 * s,
                'b1': b1 * s,
                'b2': b2 * s,
                'rho': rho * 2 - 1,  # map [0,1] → [-1,1]
            }
            comp['cov'] = np.array([[1.0, comp['rho']], [comp['rho'], 1.0]])
            self.components.append(comp)

        # Mixture weights (uniform if none provided)
        if weights is None:
            self.weights = np.ones(n_mix) / n_mix
        else:
            assert len(weights) == n_mix
            self.weights = np.array(weights) / np.sum(weights)

    def sample(self, ns: int = 1000):
        counts = np.random.multinomial(ns, self.weights)
        samples = []

        for k, count in enumerate(counts):
            if count <=1 :
                count = 2  # Ensure at least 2 samples per component for stability
            comp = self.components[k]
            z = multivariate_normal(mean=[0, 0], cov=comp['cov']).rvs(size=count)
            u = np.clip(norm.cdf(z), self.eps, 1 - self.eps)
            xa1 = beta(comp['a1'], comp['a2']).ppf(u[:, 0])
            xa2 = beta(comp['b1'], comp['b2']).ppf(u[:, 1])
            samples.append(np.stack([xa1, xa2], axis=1))

        return np.vstack(samples)[:ns, :]


def compute_beta_update_evidence(beta, log_likelihoods, log_evidence, prev_ESS):
    """  Computes beta for the next stage and updated model evidence
    Parameters
    ----------
        beta : float  stage parameter.
        log_likelihoods : numpy array of size N  log likelihood values at all particles
        log_evidence : float  log of evidence.
        prev_ESS : int   effective sample size of previous stage
    Returns
    -------
        new_beta : float    stage parameter for next stage.
        log_evidence : float  updated log evidence.
        Wm_n : numpy array of size N    weights of particles for the next stage
        ESS : float      effective sample size of new stage
    """
    old_beta, min_beta, max_beta = beta, beta, 2.0
    N, rN = len(log_likelihoods), max(0.95 * prev_ESS, 50)  # min particles 50

    while max_beta - min_beta > 1e-6:  # min step size
        new_beta = 0.5 * (max_beta + min_beta)
        inc_beta = new_beta - old_beta
        weights = np.exp(inc_beta * (log_likelihoods - log_likelihoods.max()))
        wm_normalized = weights / sum(weights)
        ess = int(1 / np.sum(wm_normalized ** 2))
        if ess == rN:
            break
        elif ess < rN:
            max_beta = new_beta
        else:
            min_beta = new_beta

    if new_beta >= 1:
        new_beta = 1
        inc_beta = new_beta - old_beta  # plausible weights of Sm corresponding to new beta
        weights = np.exp(inc_beta * (log_likelihoods - log_likelihoods.max()))
        wm_normalized = weights / sum(weights)
    log_evidence = log_evidence + np.log((sum(weights) / N))

    return new_beta, log_evidence, wm_normalized, ess


def get_log_prior_samples(samples, all_pars):
    """ Computes log_prior value at all particles
    s : numpy array of size N x Np (N samples x N parameters)
    all_pars:  list of PDFs objects, length Np
        all_pars[i]: is a PDF object with a callable log_pdf_eval method.
    Returns
     log_p : numpy array of size N log prior at all N particles .
    """
    log_p = 0
    for i in range(len(samples)):
        log_p = log_p + all_pars[i].log_pdf_eval(samples[i])
    return log_p


def initial_population(N, all_pars) -> np.ndarray:
    """
    Generates initial population from prior distribution
    N : float   number of particles.
    all_pars : list of size Np is number of parameters; all_pars[i] is object of type pdfs
    ini_pop : numpy array of size N x Np  initial population.
    """
    ini_pop = np.zeros((N, len(all_pars)))
    for i in range(len(all_pars)):
        ini_pop[:, i] = all_pars[i].generate_rns(N)
    return ini_pop


def prepare_all_pars(n_epistemic: int = 3, n_mix: int = 2):
    """
    Prepare prior for the NASA case study with a mixture of Beta marginals + copula correlation.
    Each mixture component gets its own (a1, a2, b1, b2, rho), scaled ∈ [0.001, 0.999].
    """
    l, u = 0.001, 0.999
    epistemic_domain = {}
    for ei in range(n_epistemic):
        epistemic_domain.update({f'xe{ei}': (l, u)})

    all_pars = []
    # Epistemic priors (uniform)
    for name, (lb, ub) in epistemic_domain.items():
        all_pars.append(PriorUniform(lb, ub, name=name))

    # Mixture components: (a1, a2, b1, b2, rho) × n_mix
    for i in range(n_mix):
        all_pars.append(PriorUniform(l, u, name=f'a1_m{i}'))
        all_pars.append(PriorUniform(l, u, name=f'a2_m{i}'))
        all_pars.append(PriorUniform(l, u, name=f'b1_m{i}'))
        all_pars.append(PriorUniform(l, u, name=f'b2_m{i}'))
        all_pars.append(PriorUniform(l, u, name=f'rho_m{i}'))

    return all_pars


def prepare_prior_beta_model_NASA():
    """ Prepare the prior for the beta model used in the NASA case study."""
    l, u = 0.001, 0.999
    epistemic_domain = {
        'xe1':  (l, u),
        'xe2':  (l, u),
        'xe3':  (l, u),
        'xa1a': (l, u), # marginal 1
        'xa2a': (l, u),
        'xa1b': (l, u), # marginal 2
        'xa2b': (l, u),
        'rho':  (l, u),  #  copula param
    }

    lb = np.array([v[0] for v in epistemic_domain.values()])
    ub = np.array([v[1] for v in epistemic_domain.values()])
    param_names = list(epistemic_domain.keys())
    all_pars = [PriorUniform(l, u, name=n) for l, u, n in zip(lb, ub, param_names)]
    return all_pars, epistemic_domain


def propose(current, covariance, n):
    """  proposal distribution for MCMC in perturbation stage"""
    return np.random.multivariate_normal(current, covariance, n)



def MCMC_MH(particle_num,
            Em,
            Nm_steps,
            current,
            likelihood_current,
            posterior_current,
            beta,
            numAccepts,
            all_pars,
            log_likelihood,
            ):
    """
    Markov chain Monte Carlo using Metropolis-Hastings
    "perturbs" each particle using MCMC-MH

    Parameters
    ----------
    particle_num : int
        particle number
    Em : numpy array of size Np x Np
        proposal covarince matrix.
    Nm_steps : int
        number of perturbation steps.
    current : numpy array of size Np
        current particle location
    likelihood_current : float
        log likelihood value at current particle
    posterior_current : float
        log posterior value at current particle
    beta : float
        stage parameter.
    numAccepts : int
        total number of accepts
    all_pars : : list of size Np
        Np is number of parameters
        all_pars[i] is object of type pdfs
        all parameters to be inferred.
    log_likelihood : function
        log likelihood function to be defined in main.py.

    Returns
    -------
    current : numpy array of size Np
        perturbed particle location
    likelihood_current : float
        log likelihood value at perturbed particle
    posterior_current : float
        log posterior value at perturbed particle
    numAccepts : int
        total number of accepts during perturbation (MCMC - MH)

    """
    all_proposals, all_PLP = [],  []
    deltas = propose(np.zeros(len(current)), Em, Nm_steps)
    print(f"MCMC_MH: running particle = {particle_num}, for steps {Nm_steps}" )

    for j2 in range(Nm_steps):
        delta = deltas[j2]
        proposal = current + delta
        prior_proposal = log_prior(proposal, all_pars)

        if not np.isfinite(prior_proposal):

            likelihood_proposal = -np.inf  # dont run the model
            posterior_proposal = -np.inf
        # proposal satisfies the prior constraints
        else:
            likelihood_proposal = log_likelihood(particle_num, proposal)
            posterior_proposal = prior_proposal + likelihood_proposal * beta

        log_acceptance = posterior_proposal - posterior_current

        all_proposals.append(proposal)
        all_PLP.append([prior_proposal, likelihood_proposal, posterior_proposal])

        if np.isfinite(log_acceptance) and (np.log(np.random.uniform()) < log_acceptance):
            current = proposal  # accept
            posterior_current = posterior_proposal
            likelihood_current = likelihood_proposal
            numAccepts += 1

    # gather all last samples
    return current, likelihood_current, posterior_current, numAccepts



from datetime import datetime
import time
import calendar 

if __name__ == '__main__':
    print(time.time())
    print(datetime.timestamp(datetime.now())) # Unix time
    print(time.gmtime())
    print(calendar.timegm(time.gmtime())) # Unix time, see https://en.wikipedia.org/wiki/Unix_time